import logging
import os
import yaml

import famegui.models as models
import fameio.source.loader as fameio


def _check_all_agents_have_display_coords(scenario: models.Scenario):
    if len(scenario.agents) == 0:
        return True
    for a in scenario.agents:
        if a.display_xy is None:
            return False
        return True


class CustomFameIOFactory:
    @staticmethod
    def new_schema_from_dict(definitions: dict) -> models.Schema:
        return models.Schema.from_dict(definitions)

    @staticmethod
    def new_general_properties_from_dict(definitions: dict) -> models.GeneralProperties:
        return models.GeneralProperties.from_dict(definitions)

    @staticmethod
    def new_agent_from_dict(definitions: dict) -> models.Agent:
        return models.Agent.from_dict(definitions)

    @staticmethod
    def new_contract_from_dict(definitions: dict) -> models.Contract:
        return models.Contract.from_dict(definitions)


class ScenarioLoader:
    @staticmethod
    def load_yaml_file(file_path: str, path_resolver: fameio.PathResolver) -> models.Scenario:
        """Load (read and parse) a YAML scenario file"""
        file_path = os.path.abspath(file_path)
        yaml_dict = fameio.load_yaml(file_path, path_resolver)
        scenario = models.Scenario.from_dict(definitions=yaml_dict, factory=CustomFameIOFactory())

        # check if layout generation is necessary
        if not _check_all_agents_have_display_coords(scenario):
            logging.info("at least one Agent does not have graph layout coords (X,Y): applying layouting to all agents")
            models.layout_agents(scenario)
        assert _check_all_agents_have_display_coords(scenario)

        return scenario

    @staticmethod
    def save_to_yaml_file(scenario: models.Scenario, file_path: str):
        """Save the given scenario to a YAML file"""
        logging.info("saving scenario to file {}".format(file_path))
        assert os.path.isabs(file_path)

        try:
            with open(file_path, "w") as f:
                yaml.dump(scenario.to_dict(), f)
        except Exception as e:
            raise RuntimeError("failed to save scenario to file '{}'".format(file_path)) from e



import logging
import typing
from typing import List

import fameio.source.scenario as fameio
import fameio.source.schema as fameio_schema
from fameio.source.scenario import Contract, Attribute

import famegui.models as models


class Scenario(fameio.Scenario):
    def __init__(self, schema: fameio_schema.Schema, props: fameio.GeneralProperties):
        super().__init__(schema, props)

    @property
    def contracts(self) -> List[Contract]:
        return super().contracts

    def del_all_related_contracts(self, agent_id: int) -> None:
        """Delete all contracts related to the given agent"""
        contracts_to_delete = []
        for idx, contract in enumerate(super().contracts[:]):
            if contract.sender_id == agent_id or contract.receiver_id == agent_id:
                contracts_to_delete.append(contract)
        for contract in contracts_to_delete:
            super().contracts.remove(contract)

    def set_agent_display_xy(self, agent_id: int, x: int, y: int) -> None:
        for a in self.agents:
            if a.id == agent_id:
                a.set_display_xy(x, y)
                return
        raise RuntimeError(f"Unknown agent with ID '{agent_id}'")

    def update_properties(self, props: models.GeneralProperties) -> bool:
        """Sync the scenario properties with the given ones, returns True if some changes"""
        if props.to_dict() != self.general_properties.to_dict():
            logging.info("Updating scenario general properties")
            self._general_props = props
            return True
        return False

    def get_amount_of_related_contracts(self, agent_id) -> int:
        """Return the amount of directly connected agents to the given agent_id"""
        amount = 0
        for contract in self.contracts:
            if contract.sender_id == agent_id or contract.receiver_id == agent_id:
                amount += 1
        return amount

    def get_amount_connected_agents(self, agent_id: int) -> int:
        """Return the amount of directly connected agents to the given agent_id"""
        agent_id_set = set()

        for contract in self.contracts:
            if contract.sender_id == agent_id:
                agent_id_set.add(contract.receiver_id)
                continue
            if contract.receiver_id == agent_id:
                agent_id_set.add(contract.sender_id)

        return len(agent_id_set)

    def get_agent_by_id(self, agent_id: int) -> models.Agent:
        """Return the agent object model with the given agent_id"""
        for agent in self.agents:
            if agent.id == agent_id:
                return agent
        raise RuntimeError(f"Unknown agent with ID '{agent_id}'")

    def get_all_related_agents_ids(self, agent_id: int) -> List[int]:
        """Return a list of all agents related to the given agent_id"""
        agent_id_set = set()

        for contract in self.contracts:
            if contract.sender_id == agent_id:
                agent_id_set.add(contract.receiver_id)
                continue
            if contract.receiver_id == agent_id:
                agent_id_set.add(contract.sender_id)

        return list(agent_id_set)

    def add_contract(self, contract: models.Contract) -> None:
        # TODO: move here the logic located in MainController._create_contract_model
        super().add_contract(contract)

    def update_agent(self, updated_agent: models.Agent) -> None:
        """Update the agent with the given updated_agent."""

        for idx, agent in enumerate(super().agents[:]):
            if agent.id == updated_agent.id:
                for attr, value in updated_agent.attributes.items():
                    if value == "" or value is None:
                        continue

                    super().agents[idx].attributes[attr] = Attribute(attr, value)

    def agent_exists(self, agent_id: int) -> bool:
        """Check if an agent with the given id exists in the scenario."""
        return any([True for agent in self.agents if agent.id == agent_id])

    def update_contract(
        self, old_contract: models.Contract, updated_contract: Contract
    ) -> None:
        """Update the contract with the given updated_contract."""
        """By searching the contract with the same product_name, sender_id 
        and receiver_id. to replace it with the updated_contract!!!"""

        # TODO: Far Future Exception: if a contract has empty required attributes

        for idx, contract in enumerate(super().contracts[:]):
            if (
                contract.product_name == old_contract.product_name
                and contract.sender_id == old_contract.sender_id
                and contract.receiver_id == old_contract.receiver_id
            ):
                super().contracts[idx] = updated_contract

    @property
    def agents(self) -> typing.List[models.Agent]:
        """Return the list of agents in the scenario."""
        return super().agents



import logging
import os

import fameio.source as fameio
import famegui.models as models


def write_protobuf_output(scenario: models.Scenario, output_path: str, path_resolver: fameio.PathResolver) -> None:
    logging.debug("generating protobuf output to '{}'".format(output_path))
    assert os.path.isabs(output_path)

    fameio.SchemaValidator.ensure_is_valid_scenario(scenario)
    writer = fameio.ProtoWriter(output_path, path_resolver)
    writer.write_validated_scenario(scenario)
    logging.info("protobuf output was generated to '{}'".format(output_path))



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'dialog_export_protobuf.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_DialogExportProtobuf(object):
    def setupUi(self, DialogExportProtobuf):
        if not DialogExportProtobuf.objectName():
            DialogExportProtobuf.setObjectName(u"DialogExportProtobuf")
        DialogExportProtobuf.resize(476, 230)
        self.verticalLayout_3 = QVBoxLayout(DialogExportProtobuf)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.horizontalLayout_6 = QHBoxLayout()
        self.horizontalLayout_6.setObjectName(u"horizontalLayout_6")
        self.groupBoxSimulation = QGroupBox(DialogExportProtobuf)
        self.groupBoxSimulation.setObjectName(u"groupBoxSimulation")
        self.groupBoxSimulation.setMinimumSize(QSize(250, 0))
        self.verticalLayout = QVBoxLayout(self.groupBoxSimulation)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.label = QLabel(self.groupBoxSimulation)
        self.label.setObjectName(u"label")
        self.label.setMinimumSize(QSize(80, 0))

        self.horizontalLayout.addWidget(self.label)

        self.lineEditStartTime = QLineEdit(self.groupBoxSimulation)
        self.lineEditStartTime.setObjectName(u"lineEditStartTime")
        self.lineEditStartTime.setMinimumSize(QSize(0, 0))

        self.horizontalLayout.addWidget(self.lineEditStartTime)

        self.verticalLayout.addLayout(self.horizontalLayout)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.label_2 = QLabel(self.groupBoxSimulation)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setMinimumSize(QSize(80, 0))

        self.horizontalLayout_2.addWidget(self.label_2)

        self.lineEditStopTime = QLineEdit(self.groupBoxSimulation)
        self.lineEditStopTime.setObjectName(u"lineEditStopTime")
        self.lineEditStopTime.setMinimumSize(QSize(0, 0))

        self.horizontalLayout_2.addWidget(self.lineEditStopTime)

        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.label_3 = QLabel(self.groupBoxSimulation)
        self.label_3.setObjectName(u"label_3")
        self.label_3.setMinimumSize(QSize(80, 0))

        self.horizontalLayout_3.addWidget(self.label_3)

        self.lineEditRandomSeed = QLineEdit(self.groupBoxSimulation)
        self.lineEditRandomSeed.setObjectName(u"lineEditRandomSeed")

        self.horizontalLayout_3.addWidget(self.lineEditRandomSeed)

        self.verticalLayout.addLayout(self.horizontalLayout_3)

        self.horizontalLayout_6.addWidget(self.groupBoxSimulation)

        self.groupBoxOutput = QGroupBox(DialogExportProtobuf)
        self.groupBoxOutput.setObjectName(u"groupBoxOutput")
        self.groupBoxOutput.setMinimumSize(QSize(200, 0))
        self.verticalLayout_2 = QVBoxLayout(self.groupBoxOutput)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.label_4 = QLabel(self.groupBoxOutput)
        self.label_4.setObjectName(u"label_4")

        self.horizontalLayout_4.addWidget(self.label_4)

        self.spinBoxInterval = QSpinBox(self.groupBoxOutput)
        self.spinBoxInterval.setObjectName(u"spinBoxInterval")

        self.horizontalLayout_4.addWidget(self.spinBoxInterval)

        self.verticalLayout_2.addLayout(self.horizontalLayout_4)

        self.horizontalLayout_5 = QHBoxLayout()
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.label_5 = QLabel(self.groupBoxOutput)
        self.label_5.setObjectName(u"label_5")

        self.horizontalLayout_5.addWidget(self.label_5)

        self.spinBoxProcess = QSpinBox(self.groupBoxOutput)
        self.spinBoxProcess.setObjectName(u"spinBoxProcess")

        self.horizontalLayout_5.addWidget(self.spinBoxProcess)

        self.verticalLayout_2.addLayout(self.horizontalLayout_5)

        self.horizontalLayout_6.addWidget(self.groupBoxOutput)

        self.verticalLayout_3.addLayout(self.horizontalLayout_6)

        self.groupBoxOutputFile = QGroupBox(DialogExportProtobuf)
        self.groupBoxOutputFile.setObjectName(u"groupBoxOutputFile")
        self.groupBoxOutputFile.setMinimumSize(QSize(0, 0))
        self.horizontalLayout_7 = QHBoxLayout(self.groupBoxOutputFile)
        self.horizontalLayout_7.setObjectName(u"horizontalLayout_7")
        self.label_6 = QLabel(self.groupBoxOutputFile)
        self.label_6.setObjectName(u"label_6")

        self.horizontalLayout_7.addWidget(self.label_6)

        self.lineEditOutputPath = QLineEdit(self.groupBoxOutputFile)
        self.lineEditOutputPath.setObjectName(u"lineEditOutputPath")

        self.horizontalLayout_7.addWidget(self.lineEditOutputPath)

        self.buttonOutputPath = QPushButton(self.groupBoxOutputFile)
        self.buttonOutputPath.setObjectName(u"buttonOutputPath")
        sizePolicy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.buttonOutputPath.sizePolicy().hasHeightForWidth()
        )
        self.buttonOutputPath.setSizePolicy(sizePolicy)
        self.buttonOutputPath.setMaximumSize(QSize(30, 16777215))

        self.horizontalLayout_7.addWidget(self.buttonOutputPath)

        self.verticalLayout_3.addWidget(self.groupBoxOutputFile)

        self.verticalSpacer = QSpacerItem(
            20, 8, QSizePolicy.Minimum, QSizePolicy.Expanding
        )

        self.verticalLayout_3.addItem(self.verticalSpacer)

        self.buttonBox = QDialogButtonBox(DialogExportProtobuf)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)

        self.verticalLayout_3.addWidget(self.buttonBox)

        self.retranslateUi(DialogExportProtobuf)
        self.buttonBox.accepted.connect(DialogExportProtobuf.accept)
        self.buttonBox.rejected.connect(DialogExportProtobuf.reject)

        QMetaObject.connectSlotsByName(DialogExportProtobuf)

    # setupUi

    def retranslateUi(self, DialogExportProtobuf):
        self.groupBoxSimulation.setTitle(
            QCoreApplication.translate("DialogExportProtobuf", u"Simulation", None)
        )
        self.label.setText(
            QCoreApplication.translate("DialogExportProtobuf", u"Start time:", None)
        )
        self.label_2.setText(
            QCoreApplication.translate("DialogExportProtobuf", u"Stop time:", None)
        )
        self.label_3.setText(
            QCoreApplication.translate("DialogExportProtobuf", u"Random seed:", None)
        )
        self.groupBoxOutput.setTitle(
            QCoreApplication.translate("DialogExportProtobuf", u"Output", None)
        )
        self.label_4.setText(
            QCoreApplication.translate("DialogExportProtobuf", u"Interval:", None)
        )
        self.label_5.setText(
            QCoreApplication.translate("DialogExportProtobuf", u"Process:", None)
        )
        self.groupBoxOutputFile.setTitle(
            QCoreApplication.translate("DialogExportProtobuf", u"Output file", None)
        )
        self.label_6.setText(
            QCoreApplication.translate(
                "DialogExportProtobuf", u"Protobuf output file:", None
            )
        )
        self.buttonOutputPath.setText(
            QCoreApplication.translate("DialogExportProtobuf", u"...", None)
        )
        pass

    # retranslateUi



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'dialog_newcontract.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_DialogNewContract(object):
    def setupUi(self, DialogNewContract):
        if not DialogNewContract.objectName():
            DialogNewContract.setObjectName(u"DialogNewContract")
        DialogNewContract.resize(600, 226)
        self.verticalLayout = QVBoxLayout(DialogNewContract)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.labelDescr = QLabel(DialogNewContract)
        self.labelDescr.setObjectName(u"labelDescr")
        self.labelDescr.setWordWrap(True)

        self.verticalLayout.addWidget(self.labelDescr)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.label = QLabel(DialogNewContract)
        self.label.setObjectName(u"label")

        self.horizontalLayout.addWidget(self.label)

        self.comboBoxProduct = QComboBox(DialogNewContract)
        self.comboBoxProduct.setObjectName(u"comboBoxProduct")

        self.horizontalLayout.addWidget(self.comboBoxProduct)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayoutFirstDeliveryTime = QHBoxLayout()
        self.horizontalLayoutFirstDeliveryTime.setObjectName(u"horizontalLayoutFirstDeliveryTime")
        self.label_3 = QLabel(DialogNewContract)
        self.label_3.setObjectName(u"label_3")

        self.horizontalLayoutFirstDeliveryTime.addWidget(self.label_3)

        self.lineFirstDeliveryNonFameTime = QDateTimeEdit(DialogNewContract)
        self.lineFirstDeliveryNonFameTime.setObjectName(u"lineFirstDeliveryNonFameTime")
        self.lineFirstDeliveryNonFameTime.setMinimumSize(QSize(300, 16777215))

        self.horizontalLayoutFirstDeliveryTime.addWidget(self.lineFirstDeliveryNonFameTime)

        self.lineFirstDeliveryTime = QLineEdit(DialogNewContract)
        self.lineFirstDeliveryTime.setObjectName(u"lineFirstDeliveryTime")

        self.horizontalLayoutFirstDeliveryTime.addWidget(self.lineFirstDeliveryTime)


        self.horizontalLayout_2.addLayout(self.horizontalLayoutFirstDeliveryTime)


        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.label_4 = QLabel(DialogNewContract)
        self.label_4.setObjectName(u"label_4")

        self.horizontalLayout_3.addWidget(self.label_4)

        self.lineDeliveryInterval = QLineEdit(DialogNewContract)
        self.lineDeliveryInterval.setObjectName(u"lineDeliveryInterval")
        self.lineDeliveryInterval.setMaximumSize(QSize(160, 16777215))

        self.horizontalLayout_3.addWidget(self.lineDeliveryInterval)


        self.verticalLayout.addLayout(self.horizontalLayout_3)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.label_5 = QLabel(DialogNewContract)
        self.label_5.setObjectName(u"label_5")

        self.horizontalLayout_4.addWidget(self.label_5)

        self.lineExpirationTimeNonFameTime = QDateTimeEdit(DialogNewContract)
        self.lineExpirationTimeNonFameTime.setObjectName(u"lineExpirationTimeNonFameTime")
        self.lineExpirationTimeNonFameTime.setMinimumSize(QSize(300, 16777215))

        self.horizontalLayout_4.addWidget(self.lineExpirationTimeNonFameTime)

        self.lineExpirationTime = QLineEdit(DialogNewContract)
        self.lineExpirationTime.setObjectName(u"lineExpirationTime")

        self.horizontalLayout_4.addWidget(self.lineExpirationTime)


        self.verticalLayout.addLayout(self.horizontalLayout_4)

        self.buttonBox = QDialogButtonBox(DialogNewContract)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.buttonBox)


        self.retranslateUi(DialogNewContract)
        self.buttonBox.accepted.connect(DialogNewContract.accept)
        self.buttonBox.rejected.connect(DialogNewContract.reject)

        QMetaObject.connectSlotsByName(DialogNewContract)
    # setupUi

    def retranslateUi(self, DialogNewContract):
        DialogNewContract.setWindowTitle(QCoreApplication.translate("DialogNewContract", u"Dialog", None))
        self.labelDescr.setText(QCoreApplication.translate("DialogNewContract", u"<html><head/><body><p>Details of the <span style=\"\n"
"                            font-weight:600;\">new contract</span> between agent #XXX and agent #YYY:</p></body></html>\n"
"                        ", None))
        self.label.setText(QCoreApplication.translate("DialogNewContract", u"Product:", None))
        self.label_3.setText(QCoreApplication.translate("DialogNewContract", u"First delivery time:", None))
        self.label_4.setText(QCoreApplication.translate("DialogNewContract", u"Delivery interval in steps:", None))
        self.label_5.setText(QCoreApplication.translate("DialogNewContract", u"Expiration time:", None))
    # retranslateUi




import typing

import PySide2
from PySide2 import QtCore, QtGui, QtWidgets
from PySide2.QtGui import QBrush, QColor
from PySide2.QtWidgets import QGraphicsItem

import weakref
import math
import logging


class ContractGraphItem(QtWidgets.QGraphicsItem):
    """Graphical representation of a contract between two agents.
    Note: do not register listeners here, instead use -> scenario_graph_view.py"""

    Type = QGraphicsItem.UserType + 1
    _highlight = False
    _single_highlight = False

    def __init__(
        self, sourceNode, destNode
    ):  # sourceNode: AgentGraphItem (sender) , destNode: AgentGraphItem (receiver)
        QGraphicsItem.__init__(self)
        self.setAcceptedMouseButtons(QtCore.Qt.NoButton)
        self._arrow_size = 10.0
        self._draw_line = QtCore.QLineF()
        self.source = weakref.ref(sourceNode)
        self.dest = weakref.ref(destNode)
        self.source().addEdge(self)
        self.dest().addEdge(self)
        self.adjust()

    def type(self):
        """override QGraphicsItem.type()"""
        return ContractGraphItem.Type

    """-> AgentGraphItem:"""

    def sourceNode(self):
        return self.source()

    def setSourceNode(self, node):
        if node is None:
            self.source = None
            return
        self.source = weakref.ref(node)
        self.adjust()

    """-> AgentGraphItem:"""

    def destNode(self):
        return self.dest()

    def setDestNode(self, node):
        if node is None:
            self.dest = None
            return
        self.dest = weakref.ref(node)
        self.adjust()

    def set_highlight_mode(self, highlight_enabled):
        if not highlight_enabled:
            self._single_highlight = False
        self._highlight = highlight_enabled

    def set_single_highlight_mode(self, highlight_enabled):
        self._single_highlight = highlight_enabled

    def adjust(self):
        # reset drawing
        self._draw_line = QtCore.QLineF()
        if not self.source() or not self.dest():
            return
        radius = 50
        src_center = self.mapFromItem(self.source(), radius, radius)
        dest_center = self.mapFromItem(self.dest(), radius, radius)

        # compute the distance between the two items
        [src_x, src_y] = src_center.toTuple()
        [dest_x, dest_y] = dest_center.toTuple()
        distance = math.sqrt((src_x - dest_x) ** 2 + (src_y - dest_y) ** 2)
        if distance <= radius * 2:
            logging.warning("items are too close to draw a line")
            return
        distance_x = dest_x - src_x
        distance_y = dest_y - src_y

        offset_ratio = radius * 1.0 / distance
        offset_x = distance_x * offset_ratio
        offset_y = distance_y * offset_ratio

        centerOffset = QtCore.QPointF(offset_x, offset_y)

        self.prepareGeometryChange()
        line_start = src_center + centerOffset
        line_end = dest_center - centerOffset
        self._draw_line = QtCore.QLineF(line_start, line_end)

    def is_in_highlight_mode(self):
        return self._highlight

    def boundingRect(self):
        if not self.source() or not self.dest():
            return QtCore.QRectF()

        penWidth = 1
        extra = (penWidth + self._arrow_size) / 2.0
        size = QtCore.QSizeF(
            self._draw_line.x2() - self._draw_line.x1(),
            self._draw_line.y2() - self._draw_line.y1(),
        )
        return (
            QtCore.QRectF(self._draw_line.p1(), size)
            .normalized()
            .adjusted(-extra, -extra, extra, extra)
        )

    def paint(self, painter, option, widget):
        if not self.source() or not self.dest():
            return

        # Draw the line itself
        if self._draw_line.length() == 0.0:
            return

        color = QtCore.Qt.black
        pen_stroke = 2

        if self._highlight:
            color = QtGui.QColor("#ff1100")  # Red

        if self._single_highlight:
            pen_stroke = 5

        painter.setPen(
            QtGui.QPen(
                color,
                pen_stroke,
                QtCore.Qt.SolidLine,
                QtCore.Qt.RoundCap,
                QtCore.Qt.RoundJoin,
            )
        )
        painter.drawLine(self._draw_line)

        # Draw the arrows if there's enough room.
        angle = math.acos(self._draw_line.dx() / self._draw_line.length())
        if self._draw_line.dy() >= 0:
            angle = (2.0 * math.pi) - angle

        line_end = self._draw_line.p2()
        destArrowP1 = line_end + QtCore.QPointF(
            math.sin(angle - math.pi / 3) * self._arrow_size,
            math.cos(angle - math.pi / 3) * self._arrow_size,
        )
        destArrowP2 = line_end + QtCore.QPointF(
            math.sin(angle - math.pi + math.pi / 3) * self._arrow_size,
            math.cos(angle - math.pi + math.pi / 3) * self._arrow_size,
        )

        painter.setBrush(color)
        painter.drawPolygon(
            QtGui.QPolygonF([self._draw_line.p2(), destArrowP1, destArrowP2])
        )


class AgentGraphItem(QtWidgets.QGraphicsItem):
    """A Agent node in the graph."""

    Type = QGraphicsItem.UserType + 2

    def zValue(self) -> float:
        z_value = 100
        return z_value

    # use a lambda rather than a Qt signal because we don't inherit from QObject
    position_changed = lambda self, x, y: None

    def setSelected(self, selected: bool, sender: bool = True) -> None:
        self._in_receiver_mode = sender
        super().setSelected(selected)

    def __init__(self, agent_id, label, color):
        QGraphicsItem.__init__(self)
        # public
        self.edgeList = []
        # private / read only
        self._label = "#{}".format(agent_id)
        self._color = color
        self._radius = 50
        self._agent_id = agent_id
        self._links = []
        self._highlight_border = False
        self._in_receiver_mode = True
        # customize graphics item
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)

    def type(self):
        """override QGraphicsItem.type()"""
        return AgentGraphItem.Type

    def addEdge(self, edge):
        self.edgeList.append(weakref.ref(edge))
        edge.adjust()

    def set_border_highlight(self, highlight: bool):
        self._highlight_border = highlight

    @property
    def agent_id(self) -> int:
        return self._agent_id

    def add_link(self, link):
        self._links.append(link)

    def boundingRect(self):
        return QtCore.QRectF(0, 0, self._radius * 2, self._radius * 2)

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemPositionChange:
            for edge in self.edgeList:
                if not edge():
                    continue
                edge().adjust()
            self.position_changed(self.x(), self.y())
        return QGraphicsItem.itemChange(self, change, value)

    def paint(self, painter, option, widget=None):
        """Qt paint method"""
        # main rectangle with its border
        if option.state & QtWidgets.QStyle.State_Selected:
            background_color = QtGui.QColor(self._color).darker()
            border_color = QtGui.QColor("#ff009a")
            border_width = 3
        else:
            background_color = QtGui.QColor(self._color)
            border_color = QtGui.QColor(0, 0, 0)
            border_width = 2
        if not self._in_receiver_mode:
            background_color = QtGui.QColor(self._color).darker()
            border_color = QtGui.QColor("#40ff00")
            border_width = 3

        item_rect = self.boundingRect()

        painter.setBrush(QtGui.QBrush(background_color))
        painter.setPen(QtGui.QPen(border_color, border_width))
        painter.drawEllipse(item_rect)

        # label
        font = QtGui.QFont("Arial", 14)
        font.setStyleStrategy(QtGui.QFont.ForceOutline)
        painter.setFont(font)
        painter.setPen(QtGui.QPen(QtGui.QColor(0, 0, 0), 1))
        painter.drawText(item_rect, QtCore.Qt.AlignCenter, self._label)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'dialog_scenario_properties.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_DialogScenarioProperties(object):
    def setupUi(self, DialogScenarioProperties):
        if not DialogScenarioProperties.objectName():
            DialogScenarioProperties.setObjectName(u"DialogScenarioProperties")
        DialogScenarioProperties.resize(800, 395)
        self.verticalLayout_3 = QVBoxLayout(DialogScenarioProperties)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.horizontalLayout_6 = QHBoxLayout()
        self.horizontalLayout_6.setObjectName(u"horizontalLayout_6")
        self.groupBoxSimulation = QGroupBox(DialogScenarioProperties)
        self.groupBoxSimulation.setObjectName(u"groupBoxSimulation")
        self.groupBoxSimulation.setMinimumSize(QSize(400, 0))
        self.verticalLayout = QVBoxLayout(self.groupBoxSimulation)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.groupBoxStartTime = QGroupBox(self.groupBoxSimulation)
        self.groupBoxStartTime.setObjectName(u"groupBoxStartTime")
        self.horizontalLayoutStartTime = QVBoxLayout(self.groupBoxStartTime)
        self.horizontalLayoutStartTime.setObjectName(u"horizontalLayoutStartTime")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.label = QLabel(self.groupBoxStartTime)
        self.label.setObjectName(u"label")
        self.label.setMinimumSize(QSize(80, 0))

        self.horizontalLayout.addWidget(self.label)

        self.lineEditStartTime = QDateTimeEdit(self.groupBoxStartTime)
        self.lineEditStartTime.setObjectName(u"lineEditStartTime")
        self.lineEditStartTime.setMinimumSize(QSize(0, 0))

        self.horizontalLayout.addWidget(self.lineEditStartTime)


        self.horizontalLayoutStartTime.addLayout(self.horizontalLayout)

        self.horizontalLayoutFameStartTime = QHBoxLayout()
        self.horizontalLayoutFameStartTime.setObjectName(u"horizontalLayoutFameStartTime")
        self.StartFameTimeLabel = QLabel(self.groupBoxStartTime)
        self.StartFameTimeLabel.setObjectName(u"StartFameTimeLabel")
        self.StartFameTimeLabel.setMinimumSize(QSize(0, 0))

        self.horizontalLayoutFameStartTime.addWidget(self.StartFameTimeLabel)

        self.fame_start_time = QLineEdit(self.groupBoxStartTime)
        self.fame_start_time.setObjectName(u"fame_start_time")

        self.horizontalLayoutFameStartTime.addWidget(self.fame_start_time)


        self.horizontalLayoutStartTime.addLayout(self.horizontalLayoutFameStartTime)


        self.verticalLayout.addWidget(self.groupBoxStartTime)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.groupBoxStopTime = QGroupBox(self.groupBoxSimulation)
        self.groupBoxStopTime.setObjectName(u"groupBoxStopTime")
        self.horizontalLayoutStopTime = QVBoxLayout(self.groupBoxStopTime)
        self.horizontalLayoutStopTime.setObjectName(u"horizontalLayoutStopTime")
        self.horizontalLayoutStopTime1 = QHBoxLayout()
        self.horizontalLayoutStopTime1.setObjectName(u"horizontalLayoutStopTime1")
        self.label_2 = QLabel(self.groupBoxStopTime)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setMinimumSize(QSize(80, 0))

        self.horizontalLayoutStopTime1.addWidget(self.label_2)

        self.lineEditStopTime = QDateTimeEdit(self.groupBoxStopTime)
        self.lineEditStopTime.setObjectName(u"lineEditStopTime")
        self.lineEditStopTime.setMinimumSize(QSize(0, 0))

        self.horizontalLayoutStopTime1.addWidget(self.lineEditStopTime)


        self.horizontalLayoutStopTime.addLayout(self.horizontalLayoutStopTime1)

        self.horizontalLayoutFameStopTime = QHBoxLayout()
        self.horizontalLayoutFameStopTime.setObjectName(u"horizontalLayoutFameStopTime")
        self.StopFameTimeLabel = QLabel(self.groupBoxStopTime)
        self.StopFameTimeLabel.setObjectName(u"StopFameTimeLabel")
        self.StopFameTimeLabel.setMinimumSize(QSize(0, 0))

        self.horizontalLayoutFameStopTime.addWidget(self.StopFameTimeLabel)

        self.fame_stop_time = QLineEdit(self.groupBoxStopTime)
        self.fame_stop_time.setObjectName(u"fame_stop_time")

        self.horizontalLayoutFameStopTime.addWidget(self.fame_stop_time)


        self.horizontalLayoutStopTime.addLayout(self.horizontalLayoutFameStopTime)


        self.horizontalLayout_2.addWidget(self.groupBoxStopTime)


        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.label_3 = QLabel(self.groupBoxSimulation)
        self.label_3.setObjectName(u"label_3")
        self.label_3.setMinimumSize(QSize(80, 0))

        self.horizontalLayout_3.addWidget(self.label_3)

        self.lineEditRandomSeed = QLineEdit(self.groupBoxSimulation)
        self.lineEditRandomSeed.setObjectName(u"lineEditRandomSeed")

        self.horizontalLayout_3.addWidget(self.lineEditRandomSeed)


        self.verticalLayout.addLayout(self.horizontalLayout_3)


        self.horizontalLayout_6.addWidget(self.groupBoxSimulation)

        self.groupBoxOutput = QGroupBox(DialogScenarioProperties)
        self.groupBoxOutput.setObjectName(u"groupBoxOutput")
        self.groupBoxOutput.setMinimumSize(QSize(200, 0))
        self.verticalLayout_2 = QVBoxLayout(self.groupBoxOutput)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.label_4 = QLabel(self.groupBoxOutput)
        self.label_4.setObjectName(u"label_4")

        self.horizontalLayout_4.addWidget(self.label_4)

        self.spinBoxInterval = QSpinBox(self.groupBoxOutput)
        self.spinBoxInterval.setObjectName(u"spinBoxInterval")

        self.horizontalLayout_4.addWidget(self.spinBoxInterval)


        self.verticalLayout_2.addLayout(self.horizontalLayout_4)

        self.horizontalLayout_5 = QHBoxLayout()
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.label_5 = QLabel(self.groupBoxOutput)
        self.label_5.setObjectName(u"label_5")

        self.horizontalLayout_5.addWidget(self.label_5)

        self.spinBoxProcess = QSpinBox(self.groupBoxOutput)
        self.spinBoxProcess.setObjectName(u"spinBoxProcess")

        self.horizontalLayout_5.addWidget(self.spinBoxProcess)


        self.verticalLayout_2.addLayout(self.horizontalLayout_5)


        self.horizontalLayout_6.addWidget(self.groupBoxOutput)


        self.verticalLayout_3.addLayout(self.horizontalLayout_6)

        self.groupBoxOutputFile = QGroupBox(DialogScenarioProperties)
        self.groupBoxOutputFile.setObjectName(u"groupBoxOutputFile")
        self.groupBoxOutputFile.setMinimumSize(QSize(0, 0))
        self.horizontalLayout_7 = QHBoxLayout(self.groupBoxOutputFile)
        self.horizontalLayout_7.setObjectName(u"horizontalLayout_7")
        self.label_6 = QLabel(self.groupBoxOutputFile)
        self.label_6.setObjectName(u"label_6")

        self.horizontalLayout_7.addWidget(self.label_6)

        self.lineEditOutputPath = QLineEdit(self.groupBoxOutputFile)
        self.lineEditOutputPath.setObjectName(u"lineEditOutputPath")

        self.horizontalLayout_7.addWidget(self.lineEditOutputPath)

        self.buttonOutputPath = QPushButton(self.groupBoxOutputFile)
        self.buttonOutputPath.setObjectName(u"buttonOutputPath")
        sizePolicy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.buttonOutputPath.sizePolicy().hasHeightForWidth())
        self.buttonOutputPath.setSizePolicy(sizePolicy)
        self.buttonOutputPath.setMaximumSize(QSize(30, 16777215))

        self.horizontalLayout_7.addWidget(self.buttonOutputPath)


        self.verticalLayout_3.addWidget(self.groupBoxOutputFile)

        self.verticalSpacer = QSpacerItem(20, 8, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout_3.addItem(self.verticalSpacer)

        self.groupBoxSchema = QGroupBox(DialogScenarioProperties)
        self.groupBoxSchema.setObjectName(u"groupBoxSchema")
        self.horizontalLayout_8 = QHBoxLayout(self.groupBoxSchema)
        self.horizontalLayout_8.setObjectName(u"horizontalLayout_8")
        self.comboBoxSchema = QComboBox(self.groupBoxSchema)
        self.comboBoxSchema.setObjectName(u"comboBoxSchema")

        self.horizontalLayout_8.addWidget(self.comboBoxSchema)


        self.verticalLayout_3.addWidget(self.groupBoxSchema)

        self.buttonBox = QDialogButtonBox(DialogScenarioProperties)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.verticalLayout_3.addWidget(self.buttonBox)


        self.retranslateUi(DialogScenarioProperties)
        self.buttonBox.accepted.connect(DialogScenarioProperties.accept)
        self.buttonBox.rejected.connect(DialogScenarioProperties.reject)

        QMetaObject.connectSlotsByName(DialogScenarioProperties)
    # setupUi

    def retranslateUi(self, DialogScenarioProperties):
        self.groupBoxSimulation.setTitle(QCoreApplication.translate("DialogScenarioProperties", u"Simulation", None))
        self.label.setText(QCoreApplication.translate("DialogScenarioProperties", u"Start time:", None))
        self.StartFameTimeLabel.setText(QCoreApplication.translate("DialogScenarioProperties", u"FAME Start time:", None))
        self.fame_start_time.setText(QCoreApplication.translate("DialogScenarioProperties", u"TextLabel", None))
        self.label_2.setText(QCoreApplication.translate("DialogScenarioProperties", u"Stop time:", None))
        self.StopFameTimeLabel.setText(QCoreApplication.translate("DialogScenarioProperties", u"FAME Stop time:", None))
        self.label_3.setText(QCoreApplication.translate("DialogScenarioProperties", u"Random seed:", None))
        self.groupBoxOutput.setTitle(QCoreApplication.translate("DialogScenarioProperties", u"Output", None))
        self.label_4.setText(QCoreApplication.translate("DialogScenarioProperties", u"Interval:", None))
        self.label_5.setText(QCoreApplication.translate("DialogScenarioProperties", u"Process:", None))
        self.groupBoxOutputFile.setTitle(QCoreApplication.translate("DialogScenarioProperties", u"Output file", None))
        self.label_6.setText(QCoreApplication.translate("DialogScenarioProperties", u"Protobuf output file:", None))
        self.buttonOutputPath.setText(QCoreApplication.translate("DialogScenarioProperties", u"...", None))
        self.groupBoxSchema.setTitle(QCoreApplication.translate("DialogScenarioProperties", u"Schema file", None))
        pass
    # retranslateUi




import os

from PySide2 import QtCore, QtWidgets
from PySide2.QtCore import QFile, Qt
from PySide2.QtGui import QFont
from PySide2.QtUiTools import QUiLoader
from PySide2.QtWidgets import QDialog, QTreeWidget
from PySide2.QtWidgets import QTreeWidgetItem
from fameio.source.scenario import Attribute
from fameio.source.schema import AttributeSpecs

from famegui.models import Agent, Scenario
from famegui.ui.attribute_row_item import AttributeTreeItem

from PySide2.QtCore import QFile, Slot

from famegui.ui.quick_modals import gen_quick_warning_modal


class EditAgentDialog(QDialog):
    """Dialog for editing agent attributes"""

    trigger_input_validation = QtCore.Signal()
    validation_error_signal = QtCore.Signal(bool, str)

    def __init__(self, scenario, working_dir, agents: [Agent], main_ctrl):
        super().__init__()
        ui_file = QFile(os.path.join(os.path.dirname(__file__), "edit_agent_dialog.ui"))

        ui_file.open(QFile.ReadOnly)
        self._loader = QUiLoader()
        self.validation_list = []

        self._ui = self._loader.load(ui_file)
        self.attr_top_tree_items = []
        self.validation_data = []

        self._agents = agents
        self._scenario = scenario
        self._working_dir = working_dir
        self._main_ctrl = main_ctrl
        self._schema = scenario.schema

        self.init_ui()
        self._ui.setWindowFlags(
            self.windowFlags() | Qt.WindowMaximizeButtonHint | Qt.Window
        )

        self._ui.exec_()

    def init_ui(self):
        """Initialize UI elements"""
        self._ui.treeWidget.setRootIsDecorated(False)
        self._ui.treeWidget.setColumnCount(2)
        self._ui.treeWidget.setHeaderLabels([self.tr("Attribute"), self.tr("Value")])
        self._ui.treeWidget.setColumnWidth(0, 200)
        self._ui.treeWidget.setAlternatingRowColors(True)
        self._ui.treeWidgetShowSelectedAgents.setRootIsDecorated(False)

        self._ui.treeWidgetShowSelectedAgents: QTreeWidget
        self._ui.treeWidgetShowSelectedAgents.setStyleSheet(
            "QTreeWidget::item { padding: 5px; }"
        )
        self._ui.treeWidgetShowSelectedAgents.setFixedHeight(200)
        self._ui.treeWidgetShowSelectedAgents.setColumnCount(3)
        self._ui.treeWidgetShowSelectedAgents.setHeaderLabels(
            [self.tr("selected"), self.tr("#id"), self.tr("agent type")]
        )
        self._ui.treeWidgetShowSelectedAgents.setColumnWidth(0, 200)
        self._ui.treeWidgetShowSelectedAgents.setAlternatingRowColors(True)

        self.validation_error_signal.connect(self.on_err_validation)

        self._reset_attributes()
        self._ui.agent_type_name.setText(self._agents[0].type_name)
        font = QFont()
        font.setBold(True)
        font.setPointSize(12)
        self._ui.agent_type_name.setFont(font)
        self._ui.agent_type_name.setStyleSheet("margin-bottom: 10px;")
        self._ui.buttonBox.accepted.connect(
            self._on_accept
        )  # Connect OK button to accept() method
        self._ui.buttonBox.rejected.connect(
            self._on_cancel
        )  # Connect Cancel button to reject() method

        if len(self._agents) == 1:
            preloaded_agent = self._agents[0]
            preloaded_agent: Agent

            self.init_agent_data(preloaded_agent)

    def _group_invalid_data(self):
        """Group invalid data into a string to be displayed in a modal"""
        start_str = "The Following attributes are not valid:\n"

        for item in self.validation_data:
            for item_key, item_value in item.items():
                if item_key != "msg":
                    continue
                if not item["is_valid"]:
                    start_str += "- {}\n".format(item["msg"])

        return start_str

    def on_err_validation(self, is_valid, attr_name):
        """Receive validation error signal from  a panel item and append to validation data to be checked later"""

        self.validation_data = [
            item for item in self.validation_data if item["msg"] != attr_name
        ]

        self.validation_data.append({"is_valid": is_valid, "msg": attr_name})


    def init_agent_data(self, agent: Agent):
        """Initialize agent default data by building the rows of  tree widget"""
        self._ui.treeWidget.clear()

        for attr_name, attr_spec in self._scenario.schema.agent_types[
            agent.type_name
        ].attributes.items():
            default_attr = None
            attr_spec: AttributeSpecs

            if attr_spec.has_nested_attributes:
                default_attr = agent.attributes.items()

                for item_key, item_value in agent.attributes.items():
                    if item_key == attr_name:
                        default_attr = item_value
                        break

                if not isinstance(default_attr, Attribute):
                    default_attr = None

            else:
                if attr_name in agent.attributes:
                    default_attr = agent.attributes[attr_name]

            if self.validation_error_signal is None:
                raise Exception("self.validation_error_signal is None")
                

            item = AttributeTreeItem(
                self._ui.treeWidget,
                attr_name,
                attr_spec,
                self._schema,
                self._working_dir,
                validation_signal=self.trigger_input_validation,
                validation_error_signal=self.validation_error_signal,
                default_value=default_attr,
            )

            self.attr_top_tree_items.append(item)

    def _on_cancel(self):
        """Close dialog"""
        self._ui.close()

    @staticmethod
    def _match_type_and_id(agent_type, checked_agents_ids_set: set, agent: Agent):
        """Check if agent is of given type and id is in checked_agents_ids_set"""
        return agent.type_name == agent_type and agent.id in checked_agents_ids_set

    def _save_changes(self):
        '''Save changes to scenario'''

        top_lvl_tree_agents = [
            self._ui.treeWidgetShowSelectedAgents.topLevelItem(i)
            for i in range(self._ui.treeWidgetShowSelectedAgents.topLevelItemCount())
        ]
        checked_agents = [item for item in top_lvl_tree_agents if item.checkState(0) == Qt.Checked]
        checked_agents_ids = set([int(item.text(1)) for item in checked_agents])

        self._scenario: Scenario
        matching_agents = [
            agent
            for agent in self._scenario.agents
            if self._match_type_and_id(agent.type_name, checked_agents_ids, agent)
        ]

        for agent in matching_agents:
            agent.attributes.clear()
            for item in self.attr_top_tree_items:
                item: AttributeTreeItem

                if item.attr_value is not None:
                    value = item.attr_value
                    if isinstance(value, dict):
                        value: dict
                        if not value:
                            continue

                    agent.attributes[item.attr_name] = item.attr_value

            self._scenario.update_agent(agent)

        self._main_ctrl.set_unsaved_changes(True)
        self._ui.close()

    def _on_accept(self):
        '''Apply entered changes and close dialog if all data is valid'''
        self.validation_data.clear()
        self.trigger_input_validation.emit()

        # Notify the user if there are any invalid attributes with an Alert
        for validated_item in self.validation_data:
            for item_key, item_value in validated_item.items():
                if isinstance(item_value, bool):
                    if not item_value:
                        gen_quick_warning_modal(self, "Validation error", self._group_invalid_data())

                        return

        self._save_changes()

    def _reset_attributes(self):
        """Reset UI to default state"""
        self._ui.treeWidget.clear()
        self.attr_top_tree_items.clear()
        current_agent_type = self._agents[0].type_name

        for item in self._agents:
            item: Agent
            tree_item = QTreeWidgetItem(self._ui.treeWidgetShowSelectedAgents)
            tree_item.setCheckState(0, Qt.Checked)
            tree_item.setText(1, str(item.id))
            tree_item.setText(2, item.type_name)

        for attr_name, attr_spec in self._scenario.schema.agent_types[
            current_agent_type
        ].attributes.items():
            item = AttributeTreeItem(
                self._ui.treeWidget,
                attr_name,
                attr_spec,
                self._schema,
                self._working_dir,
                validation_signal=self.trigger_input_validation,
                validation_error_signal=self.validation_error_signal,
            )
            self.attr_top_tree_items.append(item)

        self.adjust_tree_size()

    def update(self) -> None:
        """Update tree view and adjust size"""
        self.adjust_tree_size()
        super().update()

    def adjust_tree_size(self):
        """Initially  Adjust treeWidget height to fit items"""
        total_height = 0
        iterator = QtWidgets.QTreeWidgetItemIterator(self._ui.treeWidget)
        while iterator.value():
            item = iterator.value()
            row = self._ui.treeWidget.indexOfTopLevelItem(item)
            total_height += self._ui.treeWidget.sizeHintForRow(row)
            iterator += 1

        # Adjust for header height and any additional margins or spacing
        total_height += (
            self._ui.treeWidget.contentsMargins().top()
            + self._ui.treeWidget.contentsMargins().bottom()
        )

        # Set the height of the QTreeWidget
        self._ui.treeWidget.setFixedHeight(total_height)
        self.fired_once = True



import typing
import logging

import fameio.source.scenario as fameio


class Agent(fameio.Agent):
    """Extends fameio.Agent with the features required for the GUI"""

    def __init__(self, agent_id: int, type_name: str):
        super().__init__(agent_id, type_name)
        self._inputs = []
        self._outputs = []
        self._display_xy = None

    @property
    def inputs(self) -> typing.List[int]:
        return self._inputs

    def add_input(self, agent_id: int) -> None:
        self._inputs.append(agent_id)

    @property
    def outputs(self) -> typing.List[int]:
        return self._outputs

    def add_output(self, agent_id: int) -> None:
        self._outputs.append(agent_id)

    @property
    def display_xy(self) -> typing.Optional[typing.List[float]]:
        return self._display_xy

    def set_display_xy(self, x: int, y: int) -> None:
        self._display_xy = [x, y]

    @classmethod
    def from_dict(cls, definitions: dict) -> "Agent":
        result = super().from_dict(definitions)
        if (
            definitions.get("ext", None) is not None
            and definitions["ext"].get("DisplayXY", None) is not None
        ):
            xy = definitions["ext"].get("DisplayXY")
            if type(xy) is list and len(xy) == 2:
                result._display_xy = [float(xy[0]), float(xy[1])]
            else:
                logging.warning(
                    "invalid 'DisplayXY' values for agent {}: {}".format(
                        result.display_id, xy
                    )
                )
        return result

    def to_dict(self) -> dict:
        result = super().to_dict()
        if self.display_xy is not None:
            result["ext"] = {"DisplayXY": self.display_xy}
        return result



import logging
from datetime import datetime

from PySide2.QtCore import QTimer
from PySide2.QtGui import QFont
from PySide2.QtWidgets import QFormLayout, QScrollArea, QLabel, QSizePolicy

from famegui.config.style_config import FAME_CONSOLE_STYLE
from famegui.time_utils import WIDGET_DATE_FORMAT_CONSOLE


class GUIConsoleHandler(logging.StreamHandler):
    """A custom logging handler that writes to a QScrollArea to display the log messages in the GUI"""

    def __init__(self, form_layout: QFormLayout, scroll_area: QScrollArea):
        logging.StreamHandler.__init__(self)
        super().__init__()
        self._form_layout = form_layout
        self._scroll_area = scroll_area
        self._labels = []

    def emit(self, record):
        """On new log message, add a new row to the form layout and scroll to the bottom"""
        msg = self.format(record)
        current_time = datetime.now().strftime(WIDGET_DATE_FORMAT_CONSOLE)

        label_widget = QLabel(f"{current_time}: {record.levelname}: {msg}")
        label_widget.setWordWrap(True)
        label_widget.setStyleSheet(FAME_CONSOLE_STYLE)
        label_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)

        font = QFont()
        font.setPointSize(10)
        label_widget.setFont(font)

        if len(self._labels) >= 150:
            # remove the oldest row if max amount reached
            old_label = self._labels.pop(0)
            self._form_layout.removeRow(old_label)

        self._form_layout.addRow(label_widget)
        self._labels.append(label_widget)  # keep a reference to the label

        self._scroll_area.update()
        self._form_layout.update()
        self._form_layout.activate()

        QTimer.singleShot(100, self.scroll_to_bottom)

    def scroll_to_bottom(self):
        """Scroll to the bottom of the scroll area"""
        scroll_bar = self._scroll_area.verticalScrollBar()
        scroll_bar.setValue(scroll_bar.maximum())



import PySide2
from PySide2 import QtCore
from PySide2.QtCore import Qt, QPoint
from PySide2.QtGui import QCursor
from PySide2.QtWidgets import QGraphicsView


class MainGraphicsWidget(QGraphicsView):
    unselect_contracts = QtCore.Signal(name="unselect_contracts")
    agent_deletion_requested = QtCore.Signal(object)

    def __init__(self, parent=None):
        super(MainGraphicsWidget, self).__init__(parent)
        self.startPos = None

    def _get_local_widget_position(self) -> QPoint:
        """transforms the global mouse position to the local widget position"""
        mouse_position = QCursor.pos()
        local_position = self.mapFromGlobal(mouse_position)
        return local_position

    def keyPressEvent(self, event: PySide2.QtGui.QKeyEvent) -> None:
        """triggers agent deletion when delete key is pressed"""
        key = event.key()
        if key == Qt.Key_Delete:
            self.agent_deletion_requested.emit(self._get_local_widget_position())
        super().keyPressEvent(event)

    def mousePressEvent(self, event):
        """event handler for remounting the scene origin point int the canvas"""
        self.unselect_contracts.emit()
        if event.modifiers() & Qt.ShiftModifier and event.button() == Qt.LeftButton:
            self.startPos = event.pos()
        else:
            super(MainGraphicsWidget, self).mousePressEvent(event)

    def wheelEvent(self, event: PySide2.QtGui.QWheelEvent) -> None:
        """prevent from unwanted movement/ side effects when scrolling with"""
        if event.modifiers() & Qt.ShiftModifier:  # bitwise comparison necessary, due to the fact that multiple
            # modifiers can be pressed at the same time
            return
        super().wheelEvent(event)

    def mouseMoveEvent(self, event):
        """compute the difference between the current cursor position and the
        previous saved origin point"""
        if self.startPos is not None:
            delta = self.startPos - event.pos()
            transform = self.transform()
            deltaX = delta.x() / transform.m11()
            deltaY = delta.y() / transform.m22()
            self.setSceneRect(self.sceneRect().translated(deltaX, deltaY))
            self.startPos = event.pos()
        else:
            super(MainGraphicsWidget, self).mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        """resets the alternative origin point"""
        self.startPos = None
        super(MainGraphicsWidget, self).mouseReleaseEvent(event)



from typing import List

from PySide2 import QtCore, QtGui, QtWidgets
from PySide2.QtCore import QTimer
from PySide2.QtWidgets import QDialog, QDialogButtonBox, QLineEdit

from famegui import models
from famegui.agent_controller import AgentController
from famegui.config.static_ui_components import build_new_contract_dlg
from famegui.generated.ui_dialog_newcontract import Ui_DialogNewContract
from famegui.time_utils import (
    convert_fame_time_to_gui_datetime,
    convert_gui_datetime_to_fame_time,
)
from famegui.ui.quick_modals import gen_quick_warning_modal


class DialogNewContract(QDialog):
    """Dialog for creating a single contract"""

    fame_to_datetime_conversion_triggered: bool = False
    WIDGET_DATE_FORMAT = "dd/MM/yyyy hh:mm:ss"

    def _configure_line_edit_for_unsigned_int(self, line_edit: QLineEdit):
        line_edit.setText("0")
        regex_uint64 = QtCore.QRegExp("\\d{1,20}")
        line_edit.setValidator(QtGui.QRegExpValidator(regex_uint64))
        line_edit.textChanged.connect(self._update_ok_button_status)

    def _configure_line_edit_for_signed_int(self, line_edit: QLineEdit):
        line_edit.setText("0")
        regex_uint64 = QtCore.QRegExp("-?\\d{1,20}")
        line_edit.setValidator(QtGui.QRegExpValidator(regex_uint64))
        line_edit.textChanged.connect(self._update_ok_button_status)

    def _configure_line_edit_for_optional_signed_int(self, line_edit: QLineEdit):
        line_edit.setText("")
        line_edit.setPlaceholderText(self.tr("optional"))
        regex_uint64 = QtCore.QRegExp("-?\\d{0,20}")
        line_edit.setValidator(QtGui.QRegExpValidator(regex_uint64))

    def __init__(
        self,
        sender: AgentController,
        receiver: AgentController,
        schema: models.Schema,
        parent=None,
    ):
        QDialog.__init__(self, parent)
        self._ui = Ui_DialogNewContract()
        self._ui.setupUi(self)
        self._sender = sender
        self._receiver = receiver

        self.setWindowTitle(self.tr("New contract"))
        self._ui.labelDescr.setText(
            self.tr(
                self.tr(
                    self.tr(build_new_contract_dlg(self._sender, self._receiver))
                )

            ))

        # force the user to select a product except if only one is available
        if self._ui.comboBoxProduct.count() != 1:
            self._ui.comboBoxProduct.setCurrentIndex(-1)

        self._ui.comboBoxProduct.currentIndexChanged.connect(self._update_ok_button_status)
        self._update_ok_button_status()

        self._prep_input_line_edits()

        # fill possible products to select based on the sender schema agent type
        sender_type = schema.agent_type_from_name(sender.type_name)
        assert sender_type is not None
        self._ui.comboBoxProduct.addItems(list(set(sender_type.products)))

        # force the user to select a product except if only one is available
        if self._ui.comboBoxProduct.count() != 1:
            self._ui.comboBoxProduct.setCurrentIndex(-1)

        # connect
        self._ui.comboBoxProduct.currentIndexChanged.connect(
            self._update_ok_button_status
        )
        self._update_ok_button_status()

        self._ui.lineFirstDeliveryTime.textChanged.connect(
            self._update_fame_time_text_areas
        )
        self._ui.lineExpirationTime.textChanged.connect(
            self._update_fame_time_text_areas
        )  #

    def _prep_date_time_fields(self):
        """configure date time fields to use the correct format and to trigger the update of the fame time fields"""
        self._ui.lineFirstDeliveryNonFameTime.setCalendarPopup(True)
        self._ui.lineFirstDeliveryNonFameTime.setDisplayFormat(self.WIDGET_DATE_FORMAT)
        self._ui.lineFirstDeliveryNonFameTime.dateTimeChanged.connect(
            self._update_fame_times
        )

        self._ui.lineExpirationTimeNonFameTime.setCalendarPopup(True)
        self._ui.lineExpirationTimeNonFameTime.setDisplayFormat(self.WIDGET_DATE_FORMAT)
        self._ui.lineExpirationTimeNonFameTime.dateTimeChanged.connect(
            self._update_fame_times
        )

    def _prep_input_line_edits(self):
        """accept uint64 numbers as specified in protobuf schema"""
        self._configure_line_edit_for_unsigned_int(self._ui.lineDeliveryInterval)
        self._configure_line_edit_for_signed_int(self._ui.lineFirstDeliveryTime)
        self._configure_line_edit_for_optional_signed_int(self._ui.lineExpirationTime)

    def make_new_contract(self) -> models.Contract:
        expiration_time_str = self._ui.lineExpirationTime.text()
        expiration_time = int(expiration_time_str) if expiration_time_str else None

        return models.Contract(
            self._sender.id,
            self._receiver.id,
            self._ui.comboBoxProduct.currentText(),
            int(self._ui.lineDeliveryInterval.text()),
            int(self._ui.lineFirstDeliveryTime.text()),
            expiration_time,
        )

    def _update_fame_time_text_areas(self):
        self.fame_to_datetime_conversion_triggered = True

        fame_first_delivery_time: str = self._ui.lineFirstDeliveryTime.text()
        if fame_first_delivery_time == "":
            return
        gui_start_time = convert_fame_time_to_gui_datetime(
            int(fame_first_delivery_time)
        )
        self._ui.lineFirstDeliveryNonFameTime.setDateTime(gui_start_time)

        fame_expiration_time: str = self._ui.lineExpirationTime.text()
        if fame_expiration_time == "":
            return

        gui_expiration_time_time = convert_fame_time_to_gui_datetime(
            int(fame_expiration_time)
        )
        self._ui.lineExpirationTimeNonFameTime.setDateTime(gui_expiration_time_time)

    def _update_fame_times(self):
        if self.fame_to_datetime_conversion_triggered:
            self.fame_to_datetime_conversion_triggered = False
            return
        self._ui.lineExpirationTime.setText(
            str(
                convert_gui_datetime_to_fame_time(
                    self._ui.lineExpirationTimeNonFameTime.text()
                )
            )
        )
        self._ui.lineFirstDeliveryTime.setText(
            str(
                convert_gui_datetime_to_fame_time(
                    self._ui.lineFirstDeliveryNonFameTime.text()
                )
            )
        )

    def _update_ok_button_status(self):
        all_fields_ok = (
            self._ui.comboBoxProduct.currentText() != ""
            and self._ui.lineDeliveryInterval.text() != ""
            and self._ui.lineFirstDeliveryTime.text() != ""
        )
        self._ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(all_fields_ok)





# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'mainwindow.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from famegui.ui.extended_tree_viewer import ExtendedTreeWidget
from famegui.ui.main_graphics_widget import MainGraphicsWidget



class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(1386, 1000)
        MainWindow.setTabShape(QTabWidget.Triangular)
        MainWindow.setDockNestingEnabled(True)
        MainWindow.setDockOptions(QMainWindow.AllowNestedDocks | QMainWindow.AnimatedDocks | QMainWindow.VerticalTabs)
        self.actionNewProject = QAction(MainWindow)
        self.actionNewProject.setObjectName(u"actionNewProject")
        icon = QIcon()
        icon.addFile(u":/icons/new.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionNewProject.setIcon(icon)
        self.actionOpenProject = QAction(MainWindow)
        self.actionOpenProject.setObjectName(u"actionOpenProject")
        icon1 = QIcon()
        icon1.addFile(u":/icons/open.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionOpenProject.setIcon(icon1)
        self.actionSaveProject = QAction(MainWindow)
        self.actionSaveProject.setObjectName(u"actionSaveProject")
        icon2 = QIcon()
        icon2.addFile(u":/icons/save-64px.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionSaveProject.setIcon(icon2)
        self.actionExit = QAction(MainWindow)
        self.actionExit.setObjectName(u"actionExit")
        self.actionReArrangeLayout = QAction(MainWindow)
        self.actionReArrangeLayout.setObjectName(u"actionReArrangeLayout")
        self.actionGeneralProperties = QAction(MainWindow)
        self.actionGeneralProperties.setObjectName(u"actionGeneralProperties")
        icon3 = QIcon()
        icon3.addFile(u":/icons/settings.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionGeneralProperties.setIcon(icon3)
        self.actionCloseProject = QAction(MainWindow)
        self.actionCloseProject.setObjectName(u"actionCloseProject")
        self.actionMakeRunConfig = QAction(MainWindow)
        self.actionMakeRunConfig.setObjectName(u"actionMakeRunConfig")
        icon4 = QIcon()
        icon4.addFile(u":/icons/run-64px.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionMakeRunConfig.setIcon(icon4)
        self.actionSaveProjectAs = QAction(MainWindow)
        self.actionSaveProjectAs.setObjectName(u"actionSaveProjectAs")
        self.actionSchemaValidation = QAction(MainWindow)
        self.actionSchemaValidation.setObjectName(u"actionSchemaValidation")
        self.centralwidget = QSplitter(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.centralwidget.setMinimumSize(QSize(400, 60))
        self.centralwidget.setMaximumSize(QSize(16777215, 16777215))
        self.centralwidget.setBaseSize(QSize(100, 600))
        self.centralwidget.setOrientation(Qt.Vertical)
        self.frame_2 = QFrame(self.centralwidget)
        self.frame_2.setObjectName(u"frame_2")
        self.frame_2.setMinimumSize(QSize(0, 60))
        self.frame_2.setMaximumSize(QSize(16777215, 80))
        self.frame_2.setFrameShape(QFrame.StyledPanel)
        self.frame_2.setFrameShadow(QFrame.Raised)
        self.verticalLayout_2 = QVBoxLayout(self.frame_2)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.horizontalSpacer_3 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_4.addItem(self.horizontalSpacer_3)

        self.labelUserName = QLabel(self.frame_2)
        self.labelUserName.setObjectName(u"labelUserName")
        font = QFont()
        font.setPointSize(14)
        font.setBold(True)
        self.labelUserName.setFont(font)
        self.labelUserName.setTextFormat(Qt.PlainText)

        self.horizontalLayout_4.addWidget(self.labelUserName)

        self.verticalLayout_2.addLayout(self.horizontalLayout_4)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_3.addItem(self.horizontalSpacer_2)

        self.labelProjectName = QLabel(self.frame_2)
        self.labelProjectName.setObjectName(u"labelProjectName")
        font1 = QFont()
        font1.setPointSize(10)
        font1.setBold(True)
        self.labelProjectName.setFont(font1)
        self.labelProjectName.setTextFormat(Qt.PlainText)

        self.horizontalLayout_3.addWidget(self.labelProjectName)

        self.verticalLayout_2.addLayout(self.horizontalLayout_3)

        self.centralwidget.addWidget(self.frame_2)
        self.frame = QFrame(self.centralwidget)
        self.frame.setObjectName(u"frame")
        self.frame.setFrameShape(QFrame.StyledPanel)
        self.frame.setFrameShadow(QFrame.Raised)
        self.horizontalLayout_2 = QHBoxLayout(self.frame)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.graphicsView = MainGraphicsWidget(self.frame)

        self.graphicsView.setObjectName(u"graphicsView")
        self.graphicsView.setRenderHints(QPainter.Antialiasing | QPainter.TextAntialiasing)
        self.graphicsView.setViewportUpdateMode(QGraphicsView.BoundingRectViewportUpdate)

        self.horizontalLayout_2.addWidget(self.graphicsView)

        self.centralwidget.addWidget(self.frame)
        self.widget = QWidget(self.centralwidget)
        self.widget.setObjectName(u"widget")
        self.horizontalLayout = QHBoxLayout(self.widget)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayoutBottom = QVBoxLayout()
        self.verticalLayoutBottom.setObjectName(u"verticalLayoutBottom")
        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.verticalLayoutBottom.addItem(self.horizontalSpacer)

        self.horizontalLayout.addLayout(self.verticalLayoutBottom)

        self.labelZoomFactor = QLabel(self.widget)
        self.labelZoomFactor.setObjectName(u"labelZoomFactor")
        font2 = QFont()
        font2.setPointSize(10)
        self.labelZoomFactor.setFont(font2)

        self.horizontalLayout.addWidget(self.labelZoomFactor)

        self.sliderZoomFactor = QSlider(self.widget)
        self.sliderZoomFactor.setObjectName(u"sliderZoomFactor")
        self.sliderZoomFactor.setMaximumSize(QSize(200, 16777215))
        self.sliderZoomFactor.setMinimum(10)
        self.sliderZoomFactor.setMaximum(200)
        self.sliderZoomFactor.setSingleStep(10)
        self.sliderZoomFactor.setValue(100)
        self.sliderZoomFactor.setOrientation(Qt.Horizontal)

        self.horizontalLayout.addWidget(self.sliderZoomFactor)

        self.centralwidget.addWidget(self.widget)
        self.widget1 = QWidget(self.centralwidget)
        self.widget1.setObjectName(u"widget1")
        self.verticalLayoutBottomDeep = QVBoxLayout(self.widget1)
        self.verticalLayoutBottomDeep.setObjectName(u"verticalLayoutBottomDeep")
        self.verticalLayoutBottomDeep.setContentsMargins(0, 0, 0, 0)
        self.consoleLogTextEdit = QScrollArea(self.widget1)
        self.consoleLogTextEdit.setObjectName(u"consoleLogTextEdit")

        self.verticalLayoutBottomDeep.addWidget(self.consoleLogTextEdit)

        self.centralwidget.addWidget(self.widget1)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(MainWindow)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setGeometry(QRect(0, 0, 1386, 26))
        self.menuok = QMenu(self.menubar)
        self.menuok.setObjectName(u"menuok")
        self.menu_layout = QMenu(self.menubar)
        self.menu_layout.setObjectName(u"menu_layout")
        self.menu_tools = QMenu(self.menubar)
        self.menu_tools.setObjectName(u"menu_tools")
        self.menu_Edit = QMenu(self.menubar)
        self.menu_Edit.setObjectName(u"menu_Edit")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(MainWindow)
        self.statusbar.setObjectName(u"statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.toolBar = QToolBar(MainWindow)
        self.toolBar.setObjectName(u"toolBar")
        sizePolicy = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.toolBar.sizePolicy().hasHeightForWidth())
        self.toolBar.setSizePolicy(sizePolicy)
        self.toolBar.setMinimumSize(QSize(0, 0))
        self.toolBar.setMaximumSize(QSize(75, 16777215))
        self.toolBar.setMovable(False)
        self.toolBar.setIconSize(QSize(48, 48))
        self.toolBar.setFloatable(False)
        MainWindow.addToolBar(Qt.LeftToolBarArea, self.toolBar)
        self.dockAttributes = QDockWidget(MainWindow)
        self.dockAttributes.setObjectName(u"dockAttributes")
        self.dockAttributes.setFeatures(QDockWidget.DockWidgetFloatable | QDockWidget.DockWidgetMovable)
        self.dockAttributes.setAllowedAreas(Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea)
        self.dockWidgetContents = QWidget()
        self.dockWidgetContents.setObjectName(u"dockWidgetContents")
        self.verticalLayout_3 = QVBoxLayout(self.dockWidgetContents)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.treeAttributes = QTreeWidget(self.dockWidgetContents)
        __qtreewidgetitem = QTreeWidgetItem()
        __qtreewidgetitem.setText(0, u"1");
        self.treeAttributes.setHeaderItem(__qtreewidgetitem)
        self.treeAttributes.setObjectName(u"treeAttributes")
        self.treeAttributes.setMinimumSize(QSize(300, 200))
        self.treeAttributes.setMaximumSize(QSize(16777215, 16777215))

        self.verticalLayout_3.addWidget(self.treeAttributes)

        self.dockAttributes.setWidget(self.dockWidgetContents)
        MainWindow.addDockWidget(Qt.LeftDockWidgetArea, self.dockAttributes)
        self.dockProjectStructure = QDockWidget(MainWindow)
        self.dockProjectStructure.setObjectName(u"dockProjectStructure")
        self.dockProjectStructure.setFeatures(QDockWidget.DockWidgetFloatable | QDockWidget.DockWidgetMovable)
        self.dockProjectStructure.setAllowedAreas(Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea)
        self.dockWidgetContents_3 = QWidget()
        self.dockWidgetContents_3.setObjectName(u"dockWidgetContents_3")
        self.verticalLayout_5 = QVBoxLayout(self.dockWidgetContents_3)
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.lineFilterPattern = QLineEdit(self.dockWidgetContents_3)
        self.lineFilterPattern.setObjectName(u"lineFilterPattern")
        font3 = QFont()
        font3.setPointSize(9)
        self.lineFilterPattern.setFont(font3)

        self.verticalLayout_5.addWidget(self.lineFilterPattern)

        self.treeProjectStructure = ExtendedTreeWidget(self.dockWidgetContents_3)
        __qtreewidgetitem1 = QTreeWidgetItem()
        __qtreewidgetitem1.setText(0, u"1");
        self.treeProjectStructure.setHeaderItem(__qtreewidgetitem1)
        self.treeProjectStructure.setObjectName(u"treeProjectStructure")
        self.treeProjectStructure.setMinimumSize(QSize(200, 200))

        self.verticalLayout_5.addWidget(self.treeProjectStructure)

        self.dockProjectStructure.setWidget(self.dockWidgetContents_3)
        MainWindow.addDockWidget(Qt.RightDockWidgetArea, self.dockProjectStructure)

        self.menubar.addAction(self.menuok.menuAction())
        self.menubar.addAction(self.menu_Edit.menuAction())
        self.menubar.addAction(self.menu_layout.menuAction())
        self.menubar.addAction(self.menu_tools.menuAction())
        self.menuok.addAction(self.actionNewProject)
        self.menuok.addAction(self.actionOpenProject)
        self.menuok.addSeparator()
        self.menuok.addAction(self.actionSaveProject)
        self.menuok.addAction(self.actionSaveProjectAs)
        self.menuok.addAction(self.actionCloseProject)
        self.menuok.addSeparator()
        self.menuok.addAction(self.actionMakeRunConfig)
        self.menuok.addSeparator()
        self.menuok.addAction(self.actionExit)
        self.menu_layout.addAction(self.actionReArrangeLayout)
        self.menu_tools.addAction(self.actionSchemaValidation)
        self.menu_Edit.addAction(self.actionGeneralProperties)
        self.toolBar.addSeparator()
        self.toolBar.addAction(self.actionNewProject)
        self.toolBar.addAction(self.actionOpenProject)
        self.toolBar.addAction(self.actionSaveProject)
        self.toolBar.addSeparator()
        self.toolBar.addAction(self.actionGeneralProperties)
        self.toolBar.addSeparator()
        self.toolBar.addAction(self.actionMakeRunConfig)

        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)

    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle("")
        self.actionNewProject.setText(QCoreApplication.translate("MainWindow", u"&New project", None))
        # if QT_CONFIG(tooltip)
        self.actionNewProject.setToolTip(QCoreApplication.translate("MainWindow", u"Create new project", None))
        # endif // QT_CONFIG(tooltip)
        # if QT_CONFIG(shortcut)
        self.actionNewProject.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+N", None))
        # endif // QT_CONFIG(shortcut)
        self.actionOpenProject.setText(QCoreApplication.translate("MainWindow", u"&Open project", None))
        # if QT_CONFIG(tooltip)
        self.actionOpenProject.setToolTip(QCoreApplication.translate("MainWindow", u"Open an existing project", None))
        # endif // QT_CONFIG(tooltip)
        # if QT_CONFIG(shortcut)
        self.actionOpenProject.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+O", None))
        # endif // QT_CONFIG(shortcut)
        self.actionSaveProject.setText(QCoreApplication.translate("MainWindow", u"&Save project", None))
        # if QT_CONFIG(tooltip)
        self.actionSaveProject.setToolTip(QCoreApplication.translate("MainWindow", u"Save project", None))
        # endif // QT_CONFIG(tooltip)
        # if QT_CONFIG(shortcut)
        self.actionSaveProject.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+S", None))
        # endif // QT_CONFIG(shortcut)
        self.actionExit.setText(QCoreApplication.translate("MainWindow", u"Exit", None))
        # if QT_CONFIG(shortcut)
        self.actionExit.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+Q", None))
        # endif // QT_CONFIG(shortcut)
        self.actionReArrangeLayout.setText(QCoreApplication.translate("MainWindow", u"Auto Rearrange Layout", None))
        # if QT_CONFIG(shortcut)
        self.actionReArrangeLayout.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+Shift+A", None))
        # endif // QT_CONFIG(shortcut)
        self.actionGeneralProperties.setText(QCoreApplication.translate("MainWindow", u"Scenario &properties", None))
        # if QT_CONFIG(tooltip)
        self.actionGeneralProperties.setToolTip(
            QCoreApplication.translate("MainWindow", u"Edit scenario properties", None))
        # endif // QT_CONFIG(tooltip)
        # if QT_CONFIG(shortcut)
        self.actionGeneralProperties.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+P", None))
        # endif // QT_CONFIG(shortcut)
        self.actionCloseProject.setText(QCoreApplication.translate("MainWindow", u"Close project", None))
        # if QT_CONFIG(tooltip)
        self.actionCloseProject.setToolTip(QCoreApplication.translate("MainWindow", u"Close current project", None))
        # endif // QT_CONFIG(tooltip)
        self.actionMakeRunConfig.setText(QCoreApplication.translate("MainWindow", u"&Make run config", None))
        # if QT_CONFIG(tooltip)
        self.actionMakeRunConfig.setToolTip(QCoreApplication.translate("MainWindow", u"Make run config", None))
        # endif // QT_CONFIG(tooltip)
        # if QT_CONFIG(shortcut)
        self.actionMakeRunConfig.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+R", None))
        # endif // QT_CONFIG(shortcut)
        self.actionSaveProjectAs.setText(QCoreApplication.translate("MainWindow", u"Save project as...", None))
        # if QT_CONFIG(tooltip)
        self.actionSaveProjectAs.setToolTip(
            QCoreApplication.translate("MainWindow", u"Save project to a new file", None))
        # endif // QT_CONFIG(tooltip)
        self.actionSchemaValidation.setText(
            QCoreApplication.translate("MainWindow", u"Perform schema validation", None))
        # if QT_CONFIG(shortcut)
        self.actionSchemaValidation.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+Shift+V", None))
        # endif // QT_CONFIG(shortcut)
        self.labelUserName.setText(QCoreApplication.translate("MainWindow", u"User name", None))
        self.labelProjectName.setText(QCoreApplication.translate("MainWindow", u"Project name", None))
        self.labelZoomFactor.setText(QCoreApplication.translate("MainWindow", u"100 %", None))
        self.menuok.setTitle(QCoreApplication.translate("MainWindow", u"&Files", None))
        self.menu_layout.setTitle(QCoreApplication.translate("MainWindow", u"Layout", None))
        self.menu_tools.setTitle(QCoreApplication.translate("MainWindow", u"Tools", None))
        self.menu_Edit.setTitle(QCoreApplication.translate("MainWindow", u"&Edit", None))
        self.toolBar.setWindowTitle(QCoreApplication.translate("MainWindow", u"toolBar", None))
        self.dockAttributes.setWindowTitle(QCoreApplication.translate("MainWindow", u"Attributes", None))
        self.dockProjectStructure.setWindowTitle(QCoreApplication.translate("MainWindow", u"Project structure", None))
        self.lineFilterPattern.setPlaceholderText(
            QCoreApplication.translate("MainWindow", u"Filter agent type...", None))
    # retranslateUi



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'dialog_newagent.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_DialogNewAgent(object):
    def setupUi(self, DialogNewAgent):
        if not DialogNewAgent.objectName():
            DialogNewAgent.setObjectName(u"DialogNewAgent")
        DialogNewAgent.resize(668, 455)
        DialogNewAgent.setMinimumSize(QSize(300, 350))
        self.verticalLayout = QVBoxLayout(DialogNewAgent)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.label = QLabel(DialogNewAgent)
        self.label.setObjectName(u"label")

        self.horizontalLayout.addWidget(self.label)

        self.comboBoxType = QComboBox(DialogNewAgent)
        self.comboBoxType.setObjectName(u"comboBoxType")
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comboBoxType.sizePolicy().hasHeightForWidth())
        self.comboBoxType.setSizePolicy(sizePolicy)

        self.horizontalLayout.addWidget(self.comboBoxType)

        self.verticalLayout.addLayout(self.horizontalLayout)

        self.treeWidget = QTreeWidget(DialogNewAgent)
        __qtreewidgetitem = QTreeWidgetItem()
        __qtreewidgetitem.setText(0, u"1")
        self.treeWidget.setHeaderItem(__qtreewidgetitem)
        self.treeWidget.setObjectName(u"treeWidget")

        self.verticalLayout.addWidget(self.treeWidget)

        self.buttonBox = QDialogButtonBox(DialogNewAgent)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(DialogNewAgent)
        self.buttonBox.accepted.connect(DialogNewAgent.accept)
        self.buttonBox.rejected.connect(DialogNewAgent.reject)

        QMetaObject.connectSlotsByName(DialogNewAgent)

    # setupUi

    def retranslateUi(self, DialogNewAgent):
        DialogNewAgent.setWindowTitle(
            QCoreApplication.translate("DialogNewAgent", u"Dialog", None)
        )
        self.label.setText(
            QCoreApplication.translate("DialogNewAgent", u"New agent type :", None)
        )

    # retranslateUi



import logging
import math
import typing
from PySide2 import QtCore
from fameio.source.schema import AttributeSpecs

from famegui import models
from famegui.agent_controller import AgentController
from famegui.appworkingdir import AppWorkingDir
import fameio.source.validator as fameio

from famegui.models import Agent, Contract


class ProjectProperties(QtCore.QObject):
    """Class used to attach extra properties to a scenario model and signal when they change"""

    changed = QtCore.Signal()

    def __init__(self):
        super().__init__()
        self.reset()

    def reset(self, file_path=""):
        self._has_unsaved_changes = False
        self._file_path = file_path
        self._validation_errors = []
        self.changed.emit()

    @property
    def has_unsaved_changes(self) -> bool:
        return self._has_unsaved_changes

    def set_unsaved_changes(self, has_changes: bool = True):
        if self._has_unsaved_changes != has_changes:
            self._has_unsaved_changes = has_changes
            self.changed.emit()

    @property
    def file_path(self) -> str:
        return self._file_path

    def set_file_path(self, file_path: str) -> None:
        if self._file_path != file_path:
            self._file_path = file_path
            self.changed.emit()

    @property
    def is_validation_successful(self) -> bool:
        return len(self._validation_errors) == 0

    @property
    def validation_errors(self) -> typing.List[str]:
        return self._validation_errors

    def clear_validation_errors(self) -> None:
        was_invalid = not self.is_validation_successful
        self._validation_errors = []
        if was_invalid:
            logging.info("schema validation status changed to valid")
            self.changed.emit()

    def set_validation_error(self, msg: str) -> None:
        assert msg != ""
        was_valid = self.is_validation_successful
        self._validation_errors = [msg]
        if was_valid:
            logging.info("schema validation status changed to invalid")
            self.changed.emit()


class MainController(QtCore.QObject):
    selected_agent_changed = QtCore.Signal(AgentController)
    agent_added = QtCore.Signal(AgentController)
    contract_added = QtCore.Signal(
        AgentController, AgentController, models.Contract
    )  # sender, receiver, contract
    layout_rearranged = QtCore.Signal()
    agents_deleted = QtCore.Signal(list)
    contracts_deleted = QtCore.Signal(int)  # agent_id

    class AgentTypeBucket:
        """Class used to store information about a type of agent in the scenario.
        Used to build up orbitals in the graph view
        """

        def __init__(self, type_name: str, agent: Agent = None, agent_list=None):
            self.type_name = type_name
            self.agent_list = [agent]
            if agent_list is not None:
                self.agent_list = agent_list

        type_name: str
        degree_per_agent: float
        radius_scale_factor_per_agent_type: float
        agent_list: typing.List[Agent]
        contract_added = QtCore.Signal(
            AgentController, AgentController, models.Contract
        )  # sender, receiver, contract
        layout_rearranged = QtCore.Signal()
        agents_deleted = QtCore.Signal(list)
        contracts_deleted = QtCore.Signal(int)  # agent_id

        class AgentTypeBucket:
            """Class used to store information about a type of agent in the scenario.
            Used to build up orbitals in the graph view
            """

            def __init__(self, type_name: str, agent: Agent = None, agent_list=None):
                self.type_name = type_name
                self.agent_list = [agent]
                if agent_list is not None:
                    self.agent_list = agent_list

            type_name: str
            degree_per_agent: float
            radius_scale_factor_per_agent_type: float
            agent_list: typing.List[Agent]

    def __init__(self, working_dir: AppWorkingDir):
        super().__init__()
        logging.debug("initializing main controller")
        self._working_dir = working_dir
        self._scenario_model = None
        self._agents: typing.Dict[int, AgentController] = {}
        self._contracts = {}
        self._project_properties = ProjectProperties()

    def reset(self, scenario: models.Scenario = None) -> None:
        self._agents = {}
        self._contracts = {}
        self._scenario_model = scenario
        self._project_properties.reset()

    @property
    def is_open(self) -> bool:
        return self._scenario_model is not None

    @property
    def has_unsaved_changes(self) -> bool:
        return self.project_properties.has_unsaved_changes

    @property
    def project_properties(self) -> ProjectProperties:
        return self._project_properties

    @property
    def schema(self) -> models.Schema:
        return self.scenario.schema

    @property
    def scenario(self) -> models.Scenario:
        return self._scenario_model

    def update_scenario_properties(self, props: models.GeneralProperties):
        has_changed = self._scenario_model.update_properties(props)
        if has_changed:
            self._project_properties.set_unsaved_changes(True)

    @property
    def can_export_protobuf(self) -> bool:
        return (
                self.is_open
                and self.project_properties.is_validation_successful
                and not self.has_unsaved_changes
        )

    @property
    def agent_count(self) -> int:
        """get total agent count of scenario"""
        return len(self._agents)

    @property
    def agent_list(self) -> typing.List[AgentController]:
        return self._agents.values()

    def agent_from_id(self, id: int) -> AgentController:
        """get agent by id"""
        assert id in self._agents
        return self._agents[id]

    @property
    def contract_count(self) -> int:
        """get total contract count of scenario"""
        return len(self._contracts)

    def get_agent_ctrl(self, agent_id: int) -> AgentController:
        """get agent controller by agent id"""
        return self._agents[agent_id]

    def get_agent_type(self, agent_id: int) -> str:
        """get agent type by agent id"""
        return self._agents[agent_id].type_name

    def get_all_agents_by_type_name(self, agent_type_name: str) -> typing.List[Agent]:
        """filter to get all agents from scenario by agent type"""
        is_agent_from_agent_type = lambda agent: agent.type_name == agent_type_name
        return list(filter(is_agent_from_agent_type, self.scenario.agents))

    def _get_agent_bucket_list(self, agent_list: list) -> list:
        """create agent buckets by agent type"""
        agent_type_bucket_list = []

        for agent in agent_list:
            for x in agent_type_bucket_list:
                if x.type_name == agent.type_name:
                    x.agent_list.append(agent)
                    break
            else:
                agent_type_bucket_list.append(
                    self.AgentTypeBucket(agent.type_name, agent)
                )
        return agent_type_bucket_list

    def prepare_and_sort_agent_buckets(self, agent_type_bucket_list):
        """sort agent buckets by length of agent list"""
        new_agents_list = []
        for item in agent_type_bucket_list:
            new_agents_list.append(item.agent_list)
        new_agents_list.sort(key=len)
        return new_agents_list

    @staticmethod
    def _split_agent_types_into_chunks(agent_type_bucket_list):
        """split huge agent bucket into separate to make canvas look better
        create more buckets to make more smaller orbitals"""

        agent_types_to_re_gen_list = []
        chunked_list = []

        for item in agent_type_bucket_list:
            if len(item.agent_list) > 360:
                agent_types_to_re_gen_list.append(item.type_name)

                for i in range(0, len(item.agent_list), 360):
                    chunked_list.append(item.agent_list[i: i + 360])
        return agent_types_to_re_gen_list, chunked_list

    @staticmethod
    def _calc_degree_per_agent_in_agent_circle(agent_type_bucket_list):
        """calculate degree per agent in agent orbitale"""
        for item in agent_type_bucket_list:
            item.degree_per_agent = 360 / len(item.agent_list)

    @staticmethod
    def _update_graph_items(agent_ctrl_list, new_agent_x, new_agent_y, agent):
        """Updates the graph item coords of the agent with the supplied id"""
        for agent_ctrl in agent_ctrl_list:
            if agent.id.__eq__(agent_ctrl.id):
                agent_ctrl.scene_item.setPos(new_agent_x, new_agent_y)
                agent_ctrl.model_was_modified.emit()
                agent_ctrl.position_changed()
                break

    def get_amount_of_related_contract(self, agent_id: int):
        """Returns the amount of contracts related to the supplied agent"""
        amount_of_related_contract = 0
        for contract in self.scenario.contracts:
            if contract.sender_id == agent_id or contract.receiver_id == agent_id:
                amount_of_related_contract += 1
        return amount_of_related_contract

    def get_all_related_contracts(self, agent_id: int):
        """'Returns a list of all contracts related to the supplied agent"""
        related_contract_list = []
        for contract in self.scenario.contracts:
            if contract.sender_id == agent_id or contract.receiver_id == agent_id:
                related_contract_list.append(contract)
        return related_contract_list

    def delete_all_related_contracts(self, agent_id: int):
        """clear all related contracts of the supplied agent. emit signal to communicate with canvas"""
        self.contracts_deleted.emit(int(agent_id))
        self.scenario.del_all_related_contracts(int(agent_id))
        self._project_properties.set_unsaved_changes(True)

    def delete_all_related_contracts_from_scenario(self, agent_id: int):
        """clear all related contracts of the supplied agent"""
        self.scenario.del_all_related_contracts(agent_id)

    def remove_contract(self, sender_id: int, receiver_id: int, product_name: str):
        """remove contract from scenario"""
        for contract in self.scenario.contracts:
            if (
                    contract.sender_id == sender_id
                    and contract.receiver_id == receiver_id
                    and contract.product_name == product_name
            ):
                self.scenario.contracts.remove(contract)
            self._project_properties.set_unsaved_changes(True)

    def remove_agent(self, agent_id: int):
        """remove agent from scenario and delete all related contracts"""
        agent_id = int(agent_id)
        agents_to_delete = []
        self.agents_deleted.emit([agent_id])
        for agent in self.scenario.agents:
            if agent.id == agent_id:
                self.scenario.agents.remove(agent)
                agents_to_delete.append(agent)
                self.delete_all_related_contracts_from_scenario(agent_id)
        self.agents_deleted.emit([agent_id])

        self._project_properties.set_unsaved_changes(True)

    # New sub-function to handle agent positioning
    def _adjust_agent_position(self, item, default_radius, radius, margin, last_x, last_y):
        '''Adjusts the position of the agent in the scene by building a orbital model'''
        for i in range(0, len(item.agent_list)):
            agent = item.agent_list[i]
            degree = item.degree_per_agent * (i + 1)

            new_agent_x, new_agent_y, temp_rad = self._calculate_new_position(
                default_radius, degree, last_x, last_y, radius, margin
            )

            if not self.rearrange_mode:
                agent.set_display_xy(int(new_agent_x), int(new_agent_y))
                return new_agent_x, new_agent_y, temp_rad, [agent]

            self._adjust_agent_postions(self._agents, new_agent_x, new_agent_y, agent)

        return last_x, last_y, default_radius, []

    # Helper function to calculate new position
    def _calculate_new_position(self, default_radius, degree, last_x, last_y, radius, margin):
        """Helper function to calculate the new position of the agent"""

        new_agent_x = default_radius * math.cos(math.radians(degree))
        new_agent_y = default_radius * math.sin(math.radians(degree))

        distance = math.sqrt((new_agent_x - last_x) ** 2 + (new_agent_y - last_y) ** 2)
        temp_rad = default_radius

        if last_x != 0.0 and last_y != 0.0:
            while distance <= (radius * 2):
                temp_rad = temp_rad * 1.1
                new_agent_x = temp_rad * math.cos(math.radians(degree))
                new_agent_y = temp_rad * math.sin(math.radians(degree))
                distance = math.sqrt((new_agent_x - last_x) ** 2 + (new_agent_y - last_y) ** 2)

        return new_agent_x, new_agent_y, temp_rad

    def delete_all_related_agents(self, agent_id: int):
        """delete all agents that are directly connected through agents to the passed agent_id"""
        unique_agent_id_set = set()
        related_contract_list = self.get_all_related_contracts(agent_id)
        unique_agent_id_set.add(agent_id)
        for contract in related_contract_list:
            unique_agent_id_set.add(contract.sender_id)
            unique_agent_id_set.add(contract.receiver_id)

        for agent in self.scenario.agents:
            if agent.id in unique_agent_id_set:
                self.scenario.agents.remove(agent)
        self._has_unsaved_changes = True

    def _build_agent_type_bucket_list(self, agent_list: list, agent_type_bucket_list: list):
        """Builds a list that wraps the agent list into agent type buckets to build a agent orbitals model"""
        for sub_list in agent_list:
            agent_type_bucket_list.append(
                self.AgentTypeBucket(sub_list[0].type_name, agent_list=sub_list)
            )
        return agent_type_bucket_list

    def _prep_buckets_for_rearrange(self):
        """split huge agent bucket into separate to make canvas look better by creating orbitals"""
        agent_list: list = self.agent_list
        agent_type_bucket_list = self._get_agent_bucket_list(agent_list)

        new_list = []
        for item in self.prepare_and_sort_agent_buckets(agent_type_bucket_list):
            for x in item:
                new_list.append(x)
        agent_types_to_re_gen_list, chunked_list = self._split_agent_types_into_chunks(
            agent_type_bucket_list
        )

        for type_to_rm in agent_types_to_re_gen_list:
            agent_type_bucket_list = [
                value
                for value in agent_type_bucket_list
                if value.type_name != type_to_rm
            ]

        agent_type_bucket_list = self._build_agent_type_bucket_list(chunked_list, agent_type_bucket_list)

        self._calc_degree_per_agent_in_agent_circle(agent_type_bucket_list)

        return sorted(agent_type_bucket_list, key=lambda x: len(x.agent_list))

    def auto_scale_scenario_graph(
            self, agent_ctrl_list: [AgentController], rearrange_mode=False
    ):
        """initiate auto scale of scenario graph into orbital model"""
        try:
            # predefined zoom steps
            default_radius = 100.0
            radius = 50
            margin = 20
            last_x = 0.0
            last_y = 0.0
            min_distance = (float(radius) * 2) + float(margin)

            prep_buckets_for_rearrange = self._prep_buckets_for_rearrange()
            last_list = []

            for item in prep_buckets_for_rearrange:
                # expand rad for new circle
                new_radius = default_radius * 1.1
                if new_radius - default_radius < min_distance:
                    new_radius = new_radius + radius * 2 + margin
                default_radius = new_radius

                for i in range(0, len(item.agent_list)):
                    agent = item.agent_list[i]
                    degree = item.degree_per_agent
                    for counter in range(0, i):
                        degree = degree + item.degree_per_agent
                    new_agent_x = default_radius * math.cos(math.radians(degree))
                    new_agent_y = default_radius * math.sin(math.radians(degree))
                    distance = math.sqrt(
                        (new_agent_x - last_x) ** 2 + (new_agent_y - last_y) ** 2
                    )
                    temp_rad: float = default_radius

                    if last_x != 0.0 and last_y != 0.0:
                        # adjust till all spacing requirements are met
                        while distance <= (radius * 2):
                            temp_rad = temp_rad * 1.1
                            new_agent_x = temp_rad * math.cos(math.radians(degree))
                            new_agent_y = temp_rad * math.sin(math.radians(degree))
                            distance = math.sqrt(
                                (new_agent_x - last_x) ** 2
                                + (new_agent_y - last_y) ** 2
                            )

                    default_radius = temp_rad
                    last_x = new_agent_x
                    last_y = new_agent_y

                    if not rearrange_mode:
                        agent.set_display_xy(int(new_agent_x), int(new_agent_y))
                        last_list.append(agent)
                        continue

                    self._update_graph_items(
                        agent_ctrl_list, new_agent_x, new_agent_y, agent
                    )

            if not rearrange_mode:
                for a in last_list:
                    self._create_agent_controller(a)
                for c in self._scenario_model.contracts:
                    self._create_contract_model(c)

        except Exception as e:
            logging.error("failed to init the given scenario: {}".format(e))
            self.reset()
            raise

        # refresh the UI
        self._project_properties.changed.emit()

    def compute_scene_rect(self) -> QtCore.QRectF:
        rect = QtCore.QRectF(0, 0, 1000, 1000)

        if len(self.agent_list) >= 5000:
            rect = QtCore.QRectF(0, 0, 100000, 100000)

        for a in self.agent_list:
            margin = 20
            item_size = a.scene_item.boundingRect().width()
            left = a.scene_item.x() - margin
            right = a.scene_item.x() + item_size + margin
            top = a.scene_item.y() + -margin
            bottom = a.scene_item.y() + item_size + margin
            if left < rect.left():
                rect.setLeft(left)
            if right > rect.right():
                rect.setRight(right)
            if top < rect.top():
                rect.setTop(top)
            if bottom > rect.bottom():
                rect.setBottom(bottom)

        return rect

    def get_contract(self, sender_id, receiver_id, product_name) -> Contract:
        for i in self.scenario.contracts:
            i: Contract
            if i.product_name.__eq__(product_name):
                if i.sender_id.__eq__(sender_id) and i.receiver_id.__eq__(receiver_id):
                    return i

                elif i.sender_id.__eq__(receiver_id) and i.receiver_id.__eq__(
                        sender_id
                ):
                    return i

    def get_help_text_for_attr(self, agent_type_name: str, attr_name: str) -> str:
        """Extracts the help text for the given attribute name of the given agent type name"""
        for name, spec in self.schema.agent_type_from_name(
                agent_type_name
        ).attributes.items():
            spec: AttributeSpecs
            if name.__eq__(attr_name):
                return spec.help_text

    # self.schema.agent_types

    def generate_new_agent_id(self):
        new_id = len(self._agents) + 1
        # note: we don't control how ids have been generated for agents created from an external source
        # so we check for possible conflict and solve it
        if new_id in self._agents:
            for i in range(1, len(self._agents) + 2):
                if i not in self._agents:
                    new_id = i
                    break
        logging.debug("generated new agent id {}".format(new_id))
        return new_id

    # the given agent id can be 0 to clear the current selection
    def set_selected_agent_id(self, agent_id: int):
        assert self.is_open
        if agent_id not in self._agents:
            assert agent_id == 0 or agent_id is None
            self.selected_agent_changed.emit(None)
        else:
            self.selected_agent_changed.emit(self._agents[agent_id])

    def add_new_agent(self, agent_model: models.Agent, x: int, y: int):
        assert self.is_open
        agent_model.set_display_xy(x, y)
        self._create_agent_controller(agent_model)

        self._scenario_model.add_agent(agent_model)
        self._project_properties.set_unsaved_changes(True)
        self.revalidate_scenario()

        logging.info(
            "created new agent {} of type '{}'".format(
                agent_model.display_id, agent_model.type_name
            )
        )

    def _create_agent_controller(self, agent_model: models.Agent):
        assert self.is_open

        # accept to add the agent even if invalid
        agent_ctrl = AgentController(agent_model)
        self._agents[agent_ctrl.id] = agent_ctrl
        agent_ctrl.model_was_modified.connect(
            self._project_properties.set_unsaved_changes
        )

        logging.info("new agent {} added".format(agent_ctrl.display_id))
        self.agent_added.emit(agent_ctrl)

    def add_new_contract(self, contract_model: models.Contract):
        self._scenario_model.add_contract(contract_model)
        self._create_contract_model(contract_model)
        self._project_properties.set_unsaved_changes(True)
        self.revalidate_scenario()
        logging.info(
            "created new contract '{}' between {} and {}".format(
                contract_model.product_name,
                contract_model.display_sender_id,
                contract_model.display_receiver_id,
            )
        )

    def set_unsaved_changes(self, unsaved: bool):
        """set the unsaved changes flag for the current scenario"""
        self._project_properties.set_unsaved_changes(unsaved)

    def _create_contract_model(self, contract: models.Contract):
        assert self.is_open

        # validate sender / receiver are known
        if contract.sender_id not in self._agents:
            raise ValueError(
                "can't add contract '{}' because sender agent id '{}' is unknown".format(
                    contract.product_name, contract.sender_id
                )
            )

        if contract.receiver_id not in self._agents:
            raise ValueError(
                "can't add contract '{}' because receiver agent id '{}' is unknown".format(
                    contract.product_name, contract.receiver_id
                )
            )

        sender_ctrl = self._agents[contract.sender_id]
        receiver_ctrl = self._agents[contract.receiver_id]

        # connect agents
        sender_ctrl.model.add_output(contract.receiver_id)
        receiver_ctrl.model.add_input(contract.sender_id)

        self.contract_added.emit(sender_ctrl, receiver_ctrl, contract)

    def rearrange_layout(self):
        """rearrange agents in the graph view into the orbital layout"""

        self.auto_scale_scenario_graph(self.agent_list, True)

    def revalidate_scenario(self):
        assert self._scenario_model is not None

        try:
            fameio.SchemaValidator.ensure_is_valid_scenario(self._scenario_model)
            self._project_properties.clear_validation_errors()
        except fameio.ValidationException as e:
            err_msg = str(e)
            logging.warning("failed to validate the scenario: {}".format(err_msg))
            self._project_properties.set_validation_error(err_msg)

    def _init_scenario_model_with_non_auto_arrange(self):
        """initialize the scenario model from the current state of the scenario coords"""
        for a in self._scenario_model.agents:
            self._create_agent_controller(a)
        for c in self._scenario_model.contracts:
            self._create_contract_model(c)

        self.revalidate_scenario()
        return

    def _adjust_agent_postions(
            self,
            agent_ctrl_list: typing.List[AgentController],
            new_agent_x: int,
            new_agent_y: int,
            agent,
    ):
        """adjust the position of the agent in the scene"""

        for agent_ctrl in agent_ctrl_list:
            if agent.id.__eq__(agent_ctrl.id):
                agent_ctrl.scene_item.setPos(new_agent_x, new_agent_y)
                agent_ctrl.model_was_modified.emit()
                agent_ctrl.position_changed()
                break

    def init_scenario_model(self, scenario: models.Scenario, file_path: str):
        """init the scenario model and create the corresponding controllers.
        Build the graph orbital model and validate the scenario.
        """

        logging.debug("opening new scenario")
        rearrange_mode = False

        if self.is_open:
            rearrange_mode = True

        self.reset()

        try:
            self._scenario_model = scenario
            self._project_properties.reset(file_path)
            self._project_properties.set_unsaved_changes(True)
            if not rearrange_mode:
                self._init_scenario_model_with_non_auto_arrange()
                return

            try:
                last_list = []
                agent_type_bucket_list = self._prep_buckets_for_rearrange()

                default_radius = 100.0
                radius = 50
                margin = 20
                last_x = 0.0
                last_y = 0.0

                min_distance = (float(radius) * 2) + float(margin)

                for item in agent_type_bucket_list:
                    new_radius = default_radius * 1.1

                    if new_radius - default_radius < min_distance:
                        new_radius = new_radius + radius * 2 + margin
                    default_radius = new_radius

                    new_radius = default_radius * 1.1
                    if new_radius - default_radius < min_distance:
                        new_radius = new_radius + radius * 2 + margin
                    default_radius = new_radius

                    last_x, last_y, default_radius, new_last_list = self._adjust_agent_position(
                        item, default_radius, radius, margin, last_x, last_y
                    )
                    last_list.extend(new_last_list)

                if not rearrange_mode:
                    for a in last_list:
                        self._create_agent_controller(a)
                    for c in self._scenario_model.contracts:
                        self._create_contract_model(c)

            except Exception as e:
                logging.error("failed to init the given scenario: {}".format(e))
                self.reset()
                raise Exception(e)

            # refresh the UI
            self._project_properties.changed.emit()

            self.revalidate_scenario()
        except Exception as e:
            logging.error("failed to init the given scenario: {}".format(e))
            self.reset()
            raise

        # refresh the UI

        self._project_properties.changed.emit()

        self.revalidate_scenario()

    def save_to_file(self, file_path):
        assert self.is_open
        models.ScenarioLoader.save_to_yaml_file(self._scenario_model, file_path)
        # update status
        self._project_properties.set_unsaved_changes(False)
        self._project_properties.set_file_path(file_path)
        self.revalidate_scenario()



import logging
import re
import typing

import fameio.source.schema as schema
from PySide2 import QtWidgets, QtGui, QtCore
from PySide2.QtCore import QSize, Signal
from PySide2.QtGui import Qt
from PySide2.QtWidgets import QVBoxLayout, QPushButton, QLabel, QSizePolicy, QHBoxLayout
from fameio.source.scenario import Attribute
from fameio.source.schema import AttributeSpecs

from famegui.appworkingdir import AppWorkingDir
from famegui.config.static_ui_descriptions import NO_SUPPORTED_EDIT_WIDGET_TEXT
from famegui.config.style_config import get_color_for_type, get_border_for_type, ERROR_TEXT_STYLE_LABEL, \
    ERROR_TEXT_STYLE_LINE_EDIT
from famegui.ui.block_item_wrapper import BlockItemWrapper
from famegui.ui.block_type_wrapper import BlockTypeWrapper
from famegui.ui.fame_input_panels import FileChooserWidgetPanel
from famegui.ui.fame_ui_elements import QFameBoldLabel, DescriptionLabel
from famegui.ui.list_block import CustomListWidget

from famegui.ui.ui_input import set_placeholder_text
from famegui.ui.ui_input_panel_wrapper import InputPanelWrapper


class RootAttributeInputPanelWrapper(QtWidgets.QWidget):
    """Wrapper for the root level attribute input panels. It is used to toggle the visibility
    of the input panels and mange all nested input panels.
    """

    def __init__(self, row, panel, parent=None):
        super(RootAttributeInputPanelWrapper, self).__init__(parent)
        self.setAttribute(Qt.WA_StyledBackground, True)
        layout = QVBoxLayout(self)
        self.layout = layout
        self.layout.setSpacing(10)
        self.content_frame: QtWidgets.QWidget = QtWidgets.QWidget(self)

        self.row: AttributeTreeItem = row
        button = QPushButton("Collapse/Expand")
        layout.addWidget(button)
        button.clicked.connect(self.toggle_content)
        layout.addWidget(self.content_frame)

        content_layout = QVBoxLayout(self.content_frame)

        self.panel = panel

        content_layout.addWidget(self.panel)

        self.initial_width = self.width()

        self.content_frame.setVisible(True)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

    def toggle_size_adjustment(self):
        """Toggle the size adjustment of the parent widget."""
        self.adjustSize()

        self.updateGeometry()
        self.row.setSizeHint(0, QSize(-1, 60))
        self.row.setSizeHint(1, QSize(-1, 60))
        self.parent().setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        self.parent().updateGeometry()
        self.parent().adjustSize()

    def toggle_content(self):
        """Toggle the visibility of the input panel by triggering layout updates."""

        if self.content_frame.isVisible():
            self.content_frame.hide()

        else:
            self.content_frame.show()

        self.parent().updateGeometry()
        self.parent().adjustSize()
        self.parent().update()

    def height(self):
        """Get the pre-calculated height of the input panel."""
        return self.content_frame.sizeHint().height() + 50


class AttributeTreeItem(QtWidgets.QTreeWidgetItem):
    """Root Holder class for all attribute tree items.
    It is used to create a tree of attributes and their input panels to access all attributes .
    """

    def __init__(
            self,
            parent: QtWidgets.QTreeWidget,
            attr_name: str,
            attr_spec: schema.AttributeSpecs,
            schema: schema.Schema,
            working_dir: AppWorkingDir,
            validation_signal: Signal = None,
            validation_error_signal: Signal = None,
            default_value=None,
    ):
        self.state_value_dict = {}
        self._attr_name = attr_name
        self._attr_spec: AttributeSpecs = attr_spec
        self.validation_signal: Signal = validation_signal
        self.validation_error_signal: Signal = validation_error_signal
        self._working_dir = working_dir
        self._default_value = default_value
        self._attr_value = None
        self._display_error = lambda has_error: None
        self.initial_height = 0
        self.schema = schema
        self.holder_widget = None

        QtWidgets.QTreeWidgetItem.__init__(self, parent, [attr_name])

        font = self.font(0)
        if self._attr_spec.is_mandatory:
            font.setBold(True)

            tooltip = self.tr("{} (mandatory)").format(self._attr_name)
        else:
            font.setItalic(True)
            tooltip = self.tr("{} (optional)").format(self._attr_name)

        self.setFont(0, font)
        self.setToolTip(0, tooltip)
        self.ui_parent = parent
        self.panel = self._create_edit_widget()

        self.holder_widget = RootAttributeInputPanelWrapper(
            self, self.panel, self.treeWidget()
        )

        color_str = get_color_for_type(self._attr_spec)

        self.holder_widget.setStyleSheet(
            f"RootAttributeInputPanelWrapper {{  background:{color_str}; border-radius: 5px;"
            "margin-top:10px; margin-bottom:30px; margin-left: 5px; margin-right: 5px; }}"
        )

        parent.setItemWidget(self, 1, self.holder_widget)
        self.holder_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

    def get_attr_spec(self):
        """Get the root AttributeSpecs of the row."""
        return self._attr_spec

    @property
    def attr_name(self) -> str:
        """Get the attribute short name of the root attribute ."""
        return self._attr_name

    def remove_none_values_and_empty_elements(self, item):
        """Recursively remove all None values and empty elements from a dictionary or list to clean up the data
        prepared for saving to a file
        """
        if isinstance(item, dict):
            # Handle dictionary: recursively clean its key-value pairs
            new_dict = {
                k: self.remove_none_values_and_empty_elements(v)
                for k, v in item.items()
                if v is not None
            }
            return {k: v for k, v in new_dict.items() if v not in [None, {}, []]}
        elif isinstance(item, list):
            # Handle list: recursively clean its elements
            new_list = [self.remove_none_values_and_empty_elements(v) for v in item]
            return [v for v in new_list if v not in [None, {}, []]]
        else:
            # Base case: return the item as-is
            return item

    def remove_none_values(self, d):
        """Helper function to remove None values from a dictionary."""
        if not isinstance(d, dict):
            return d
        return {k: self.remove_none_values(v) for k, v in d.items() if v is not None}

    @property
    def attr_value(self) -> typing.Any:
        """Get clean attribute value from a complete row"""

        cleaned_data = self.remove_none_values_and_empty_elements(
            self._get_data_from_input()
        )

        return cleaned_data

    @property
    def validation_error(self) -> typing.Optional[str]:
        if self._attr_spec.is_mandatory and self._attr_value is None:
            return "mandatory attribute '{}' is missing".format(self.attr_name)
        return None

    _supported_list_types = [
        schema.AttributeType.INTEGER,
        schema.AttributeType.LONG,
        schema.AttributeType.DOUBLE,
    ]

    @property
    def _attr_enum_values(self) -> typing.List[str]:
        """Returns the enum values of the attached AttributeSpecs"""
        assert self._attr_spec.attr_type == schema.AttributeType.ENUM
        assert not self._attr_spec.is_list
        assert not self._attr_spec.has_nested_attributes
        return self._attr_spec.values

    def build_parent_container(self, attr_spec_inner, attr_name: str, holder_widget) -> QtWidgets.QWidget:
        """Build a parent container for a block type attribute with an n depth of nested attributes"""

        value = attr_spec_inner.help_text or ""
        desc = DescriptionLabel(text=value)

        label = QFameBoldLabel(
            text=f"{attr_name}: {attr_spec_inner.attr_type.name} "
        )
        label.setWordWrap(True)

        parent_layout = QVBoxLayout()

        parent_container = BlockTypeWrapper(
            block_type_layout=parent_layout,
            spec=attr_spec_inner,
            attr_name=attr_name,
        )

        layout_header = QVBoxLayout()
        header_widget = QtWidgets.QWidget()

        header_widget.setLayout(layout_header)
        layout_header.insertWidget(0, label)
        layout_header.insertWidget(1, desc)

        parent_layout.insertWidget(0, header_widget)

        # Wrap holder_widget with a parent container and chevron icon

        # Set the stylesheet for the specific widget using its object name

        chevron_button = QPushButton(
            "Expand/Collapse"
        )  # Using a triangle as a placeholder for the chevron icon

        chevron_button.setCheckable(True)

        # Set the initial checked state to match the initial visibility of holder_widget
        chevron_button.setChecked(True)
        chevron_button.toggled.connect(holder_widget.setVisible)
        parent_layout.addWidget(chevron_button)
        parent_layout.addWidget(holder_widget)

        return parent_container

    def build_child_sub_block_list(self, attr_spec_inner, attr_name, holder: QVBoxLayout, list_data=None):
        """Build a child list block panel for a block"""

        preset_value_list = self.get_default_value(
            attr_spec_inner, attr_spec_inner.full_name, list_data
        )

        block_list = self.build_sub_block_list(
            attr_spec_inner, preset_value_list
        )

        dynamic_list = CustomListWidget(
            block_list, attr_spec_inner, self.schema, attr_name=attr_name
        )

        dynamic_list.row_creation_requested.connect(self._add_list_row)

        holder.set_widget(dynamic_list)

    def build_primitive_block_child(self, attr_spec_inner: AttributeSpecs, attr_name: str, holder: QVBoxLayout,
                                    list_data=None):
        """Build a primitive child input panel for a block"""
        default_row = QFameBoldLabel(
            text=f"{attr_name}: {attr_spec_inner.attr_type.name}"
        )

        default_row.setWordWrap(True)
        value = attr_spec_inner.help_text or ""

        desc = DescriptionLabel(text=value)

        parent_widget = BlockItemWrapper(spec=attr_spec_inner, attr_name=attr_name)

        parent_widget.add_widget(default_row)
        parent_widget.add_widget(desc)
        prim_widget = self.get_primitive_input_panel(attr_spec_inner, list_data)

        uniq_widget = f"myUniqueWidget-{id(parent_widget)}"

        parent_widget.setObjectName(uniq_widget)

        color_str = get_color_for_type(attr_spec_inner)

        border_style = get_border_for_type(attr_spec_inner)

        parent_widget.setStyleSheet(
            f"#{uniq_widget} {{"
            f"    background-color: {color_str};"
            f"    {border_style}"
            "    border-radius: 5px;"
            "}"
        )

        parent_widget.add_input_widget(prim_widget)

        holder.addWidget(parent_widget)

    def build_sub_block_list(self, attr_data: AttributeSpecs, list_data=None):
        """Build a list of sub blocks for a block type attribute with an n depth of nested attributes"""
        holder = QVBoxLayout()

        for item, attr_spec_inner in attr_data.nested_attributes.items():
            attr_spec_inner: AttributeSpecs

            if attr_spec_inner.attr_type == schema.AttributeType.BLOCK:
                holder_widget = self.build_sub_block_list(attr_spec_inner, list_data)

                parent_container = self.build_parent_container(
                    attr_spec_inner, item, holder_widget
                )

                input_wrapper = InputPanelWrapper(
                    inner_widget=parent_container,
                    signal_to_connect=self.validation_signal,
                    validation_error_signal=self.validation_error_signal,
                    attr_name=item,
                    attribute=attr_spec_inner,
                )

                holder.addWidget(input_wrapper)
                continue

            if attr_spec_inner.is_list:
                self.build_child_sub_block_list(attr_spec_inner, item, holder, list_data)

                continue

            self.build_primitive_block_child(attr_spec_inner, item, holder, list_data)

            ## here

        default_data_dict = self.get_default_value(
            attr_data, self._attr_spec.full_name, list_data
        )

        holder_wrapper = BlockTypeWrapper(
            block_type_layout=holder,
            spec=attr_data,
            attr_name=self.attr_name,
            dict_data=default_data_dict,
        )

        return holder_wrapper

    def get_list_capable_int_or_long_input_panel(self, spec: AttributeSpecs, nested_list):
        preset_value = str(
            self.get_default_value(spec, spec.full_name, nested_list)
        )

        line_edit = self._create_line_edit("0", None)
        line_edit.setText(preset_value)
        line_edit.setValidator(QtGui.QIntValidator())

        input_wrapper = InputPanelWrapper(
            inner_widget=line_edit,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
        )
        self.state_value_dict = {spec.full_name: input_wrapper}
        return input_wrapper

    def get_list_capable_time_series_input_panel(self, spec: AttributeSpecs, nested_list):
        preset_value = self.get_default_value(spec, spec.full_name, nested_list)

        file_chooser = FileChooserWidgetPanel(
            self._working_dir,
            spec.attr_type.name,
            preset_value,
            self.ui_parent,
        )

        self._display_error = file_chooser.display_error

        input_wrapper = InputPanelWrapper(
            inner_widget=file_chooser,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
        )
        self.state_value_dict = {spec.full_name: input_wrapper}
        return input_wrapper

    def get_list_capable_enum_input_panel(self, spec: AttributeSpecs, nested_list):
        combo_box = QtWidgets.QComboBox()
        enum_values = spec.values

        combo_box.addItems(enum_values)
        if len(enum_values) == 1:
            combo_box.setCurrentIndex(0)
            self._attr_value = enum_values[0]
        else:
            combo_box.setCurrentIndex(-1)

        preset_value = self.get_default_value(spec, spec.full_name, nested_list)

        set_placeholder_text(combo_box, preset_value, enum_values)

        input_wrapper = InputPanelWrapper(
            inner_widget=combo_box,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
        )
        self.state_value_dict = {spec.full_name: input_wrapper}
        return input_wrapper

    def get_list_capable_double_input_panel(self, spec: AttributeSpecs, nested_list):
        preset_value = str(
            self.get_default_value(spec, spec.full_name, nested_list)
        )

        line_edit = self._create_line_edit(preset_value, None)

        line_edit.setText(preset_value)
        validator = QtGui.QDoubleValidator()
        # accept '.' as decimal separator
        validator.setLocale(QtCore.QLocale.English)
        line_edit.setValidator(validator)

        input_wrapper = InputPanelWrapper(
            inner_widget=line_edit,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
        )

        self.state_value_dict = {spec.full_name: input_wrapper}
        return input_wrapper

    def get_primitive_input_panel(self, spec, nested_list):
        """Get a primitive input panel for a given attribute spec"""
        if spec.attr_type == schema.AttributeType.ENUM:

            return self.get_list_capable_enum_input_panel(spec, nested_list)


        elif spec.attr_type == schema.AttributeType.TIME_SERIES:

            return self.get_list_capable_time_series_input_panel(spec, nested_list)

        elif (
                spec.attr_type == schema.AttributeType.INTEGER
                or spec.attr_type == schema.AttributeType.LONG
        ):

            return self.get_list_capable_int_or_long_input_panel(spec, nested_list)

        elif spec.attr_type == schema.AttributeType.DOUBLE:

            return self.get_list_capable_double_input_panel(spec, nested_list)

        elif spec.attr_type == schema.AttributeType.BLOCK:

            return self.build_scroll_block_edit_widget(spec, nested_list)

    def _add_list_row(self, spec: AttributeSpecs, widget: QtWidgets.QWidget):
        """Add a new row to a list after the user clicked the add button"""
        widget: CustomListWidget
        block_list = self.build_sub_block_list(spec)

        input_wrapper = InputPanelWrapper(
            inner_widget=block_list,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
        )

        widget.set_widget(input_wrapper)

        self.ui_parent.update()

    def _create_edit_widget(self) -> QtWidgets.QWidget:
        """Create a input panel edit widget for the attribute"""
        spec = self._attr_spec

        if spec.is_list:
            return self.build_list_edit_widget(spec)

        if spec.attr_type == schema.AttributeType.ENUM:

            return self.build_enum_chooser_input_panel(spec)

        elif spec.attr_type == schema.AttributeType.TIME_SERIES:

            return self.build_time_series_edit_widget(spec)

        # INT INPUT PANEL

        elif (
                spec.attr_type == schema.AttributeType.INTEGER
                or spec.attr_type == schema.AttributeType.LONG
        ):
            return self.build_int_or_long_edit_widget(spec)

        # DOUBLE  INPUT PANEL
        elif spec.attr_type == schema.AttributeType.DOUBLE:

            return self.build_double_edit_widget(spec)

        elif spec.attr_type == schema.AttributeType.BLOCK:

            return self.build_block_edit_widget(spec)

    def build_block_edit_widget(self, spec: AttributeSpecs):
        block_list = self.build_sub_block_list(spec)

        input_wrapper = InputPanelWrapper(
            inner_widget=block_list,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
        )
        self.state_value_dict = {spec.full_name: input_wrapper}
        return input_wrapper

    def build_double_edit_widget(self, spec: AttributeSpecs):
        preset_value = self.get_default_value(spec, spec.full_name)

        line_edit = self._create_line_edit(preset_value, None)

        line_edit.setText(preset_value)
        validator = QtGui.QDoubleValidator()
        # accept '.' as decimal separator
        validator.setLocale(QtCore.QLocale.English)
        line_edit.setValidator(validator)

        input_wrapper = InputPanelWrapper(
            inner_widget=line_edit,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
        )
        self.state_value_dict = {spec.full_name: input_wrapper}

        return input_wrapper

    def build_enum_chooser_input_panel(self, spec: AttributeSpecs):
        combo_box = QtWidgets.QComboBox()
        enum_values = self._attr_enum_values
        combo_box.addItems(enum_values)
        if len(enum_values) == 1:
            combo_box.setCurrentIndex(0)
            self._attr_value = enum_values[0]
        else:
            combo_box.setCurrentIndex(-1)

        preset_value = self.get_default_value(
            spec,
            spec.full_name,
        )

        set_placeholder_text(combo_box, preset_value, enum_values)

        input_wrapper = InputPanelWrapper(
            inner_widget=combo_box,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
        )
        self.state_value_dict = {spec.full_name: input_wrapper}

        return input_wrapper

    def build_scroll_block_edit_widget(self, spec: AttributeSpecs, nested_list):
        scroll_area = QtWidgets.QScrollArea()
        scroll_area.setWidgetResizable(True)
        block_list = self.build_sub_block_list(spec, nested_list)

        scroll_area.setLayout(block_list)

        input_wrapper = InputPanelWrapper(
            inner_widget=block_list,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
        )
        self.state_value_dict = {spec.full_name: input_wrapper}
        return input_wrapper

    def build_time_series_edit_widget(self, spec: AttributeSpecs):
        preset_value = self.get_default_value(spec, spec.full_name)

        file_chooser = FileChooserWidgetPanel(
            self._working_dir, spec.attr_type.name, preset_value, self.treeWidget()
        )

        self._display_error = file_chooser.display_error

        input_wrapper = InputPanelWrapper(
            inner_widget=file_chooser,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
        )
        self.state_value_dict = {spec.full_name: input_wrapper}
        return input_wrapper

    def build_int_or_long_edit_widget(self, spec: AttributeSpecs):
        line_edit = self._create_line_edit("0", None)
        preset_value = self.get_default_value(spec, spec.full_name)
        line_edit.setText(preset_value)
        line_edit.setValidator(QtGui.QIntValidator())

        input_wrapper = InputPanelWrapper(
            inner_widget=line_edit,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
        )
        self.state_value_dict = {spec.full_name: input_wrapper}
        return input_wrapper

    def build_list_edit_widget(self, spec: AttributeSpecs):
        preset_value = self.get_nested_list_data()

        if preset_value is None:
            preset_value = []

        for item, idx in zip(preset_value, range(len(preset_value))):
            item: dict
            for k, v in item.items():
                v: Attribute

        if isinstance(preset_value, list):
            dynamic_list = None

            for item, idx in zip(preset_value, range(len(preset_value))):
                if idx == 0:
                    block_list = self.build_sub_block_list(spec, [item])

                    dynamic_list = CustomListWidget(
                        block_list, spec, self.schema, self.attr_name
                    )
                    continue

                item: dict

                block_list = self.build_sub_block_list(spec, [item])

                input_wrapper = InputPanelWrapper(
                    inner_widget=block_list,
                    validation_error_signal=self.validation_error_signal,
                    signal_to_connect=self.validation_signal,
                    attribute=spec,
                )

                dynamic_list.set_widget(input_wrapper)

            if dynamic_list is None:
                block_list = self.build_sub_block_list(spec)

                dynamic_list = CustomListWidget(
                    block_list, spec, self.schema, self.attr_name
                )

            dynamic_list.row_creation_requested.connect(self._add_list_row)

            input_wrapper = InputPanelWrapper(
                inner_widget=dynamic_list,
                validation_error_signal=self.validation_error_signal,
                signal_to_connect=self.validation_signal,
                attribute=spec,
            )

            self.state_value_dict = {spec.full_name: input_wrapper}

            return input_wrapper

        block_list = self.build_sub_block_list(spec, preset_value)

        dynamic_list = CustomListWidget(
            block_list, spec, self.schema, self.attr_name
        )

        dynamic_list.row_creation_requested.connect(self._add_list_row)

        input_wrapper = InputPanelWrapper(
            inner_widget=dynamic_list,
            validation_error_signal=self.validation_error_signal,
            signal_to_connect=self.validation_signal,
            attribute=spec,
        )
        self.state_value_dict = {spec.full_name: input_wrapper}
        return input_wrapper

    def search_attr_recursively(self, attr: Attribute, attr_full_name: str):
        """Search for a specific and primitive  nested attribute recursively"""
        for nested_key, nested_item in attr.nested.items():
            nested_item: Attribute

            if attr_full_name == str(nested_key):
                return nested_item

            if nested_item.has_nested:
                temp_result = self.search_attr_recursively(nested_item, attr_full_name)
                if temp_result:
                    return temp_result
                continue

        return None

    def recursive_search(self, data, key):
        """Search for a specific and primitive  nested attribute recursively for any data type"""
        if isinstance(data, list):
            for item in data:
                result = self.recursive_search(item, key)
                if result:
                    return result
        elif isinstance(data, dict):
            if key in data:
                return data[key]
            for k, v in data.items():
                result = self.recursive_search(v, key)
                v: Attribute
                if v.has_nested:
                    result = self.recursive_search(v.nested, key)

                if result:
                    return result
        return None

    def get_nested_list_data(self):
        """Get the nested list data from the root attribute spec
        Since one row can have multiple nested lists,
        we need to get the data from the root attribute spec and then search for the specific nested list data recursively
        """
        if self._default_value:
            if self._default_value.has_nested_list:
                nested_list = self._default_value.nested_list

                return nested_list
        return None

    def get_default_value(
            self, spec: AttributeSpecs, attr_full_name, nested_list_data=None
    ):
        """Get the saved value for a given attribute spec ( if it exists)"""
        self._default_value: Attribute

        if nested_list_data:
            matches = re.findall(r"(?<=\.)[^.]*$", attr_full_name)

            attr_short_name = matches[0]

            result = self.recursive_search(nested_list_data, attr_short_name)
            if result:
                return str(result.value)

        if self._default_value:
            if self._default_value.has_nested_list:
                matches = re.findall(r"(?<=\.)[^.]*$", attr_full_name)

                attr_short_name = matches[0]

                nested_list = self._default_value.nested_list

                result = self.recursive_search(nested_list, attr_short_name)
                if result:
                    if result.has_value:
                        return str(result.value)

                return None

        matches = re.findall(r"(?<=\.)[^.]*$", attr_full_name)

        attr_short_name = matches[0]

        if self._default_value:
            if self._default_value.has_nested:
                search_attr_recursively = self.search_attr_recursively(
                    self._default_value, attr_short_name
                )

                if search_attr_recursively:
                    if search_attr_recursively.has_value:
                        return str(search_attr_recursively.value)

        spec_attr_fallback_value = ""

        if not self._default_value:
            return spec_attr_fallback_value

        default_value = ""

        self._default_value: Attribute

        if self._default_value.has_nested:
            for nested_key, nested_item in self._default_value.nested.items():
                nested_item: Attribute

                if attr_full_name.__contains__(str(nested_key)):
                    if nested_item.has_value:
                        default_value = nested_item.value

                    return str(default_value)

        if self._default_value.has_value:
            default_value = str(self._default_value.value)

        default_value = (
            default_value if default_value is not None else spec_attr_fallback_value
        )

        return str(default_value)

    def _create_line_edit(self, placeholder_text: str, handler) -> QtWidgets.QLineEdit:
        """Create a line edit widget for the attribute"""
        line_edit = QtWidgets.QLineEdit()
        line_edit.setText(placeholder_text)
        self._display_error = (
            lambda has_error: line_edit.setStyleSheet(
                ERROR_TEXT_STYLE_LINE_EDIT
            )
            if has_error
            else line_edit.setStyleSheet("")
        )
        return line_edit

    def _create_unsupported_edit_widget(self, type_name: str) -> QtWidgets.QLabel:
        """Create a label widget for an unsupported attribute type (Leave for future implementation for new kinds of
        attributes)"""

        logging.error(
            NO_SUPPORTED_EDIT_WIDGET_TEXT.format(
                type_name
            )
        )
        label = QtWidgets.QLabel()
        label.setText(
            self.tr(NO_SUPPORTED_EDIT_WIDGET_TEXT.format(type_name))
        )
        label.setStyleSheet(ERROR_TEXT_STYLE_LABEL)
        return label

    def _get_data_from_input(self):
        """Get the data from the input panel"""

        attr_data = self.panel.get_input_value()

        return attr_data

    def tr(self, msg: str) -> str:
        return QtCore.QCoreApplication.translate("AttributeTreeItem", msg)



import re
import typing

import PySide2
from PySide2 import QtCore
from PySide2.QtCore import Qt
from PySide2.QtWidgets import QTreeWidget, QTreeWidgetItem

from famegui.utils import get_product_name_from_contract


class ExtendedTreeWidget(QTreeWidget):
    """Tree widget that uses custom event signals/triggers to communicate with the main application"""

    all_agents_deletion_requested = QtCore.Signal(str, name="del_all_agents_of_type")
    single_agent_deletion_requested = QtCore.Signal(int, name="del_single_agent")
    single_contract_deletion_requested = QtCore.Signal(
        int, int, str, name="del_single_agent"
    )  # agent_sender_id,agent_receiver_id, product_name

    agent_deletion_requested = QtCore.Signal(int, name="agent_deletion_requested")

    def __init__(
            self, parent: typing.Optional[PySide2.QtWidgets.QWidget] = ...
    ) -> None:
        super().__init__(parent)

    def _get_id_from_item_desc(self, text: str):
        # extract id from text
        pattern = r"#(\d+)"
        match = re.search(pattern, text)

        if match:
            result = match.group(1)
            return result
        else:
            return ""

    def process_contract_deletion(
            self, text_of_selected_item: str, current_item, agent_id_ref
    ):
        selected_agent_id = current_item.parent().data(0, QtCore.Qt.UserRole)
        connection_type = current_item.data(1, QtCore.Qt.UserRole)  # sender or receiver
        operator_name_two = get_product_name_from_contract(text_of_selected_item)

        if connection_type == "sender":
            self.single_contract_deletion_requested.emit(
                int(selected_agent_id),
                int(agent_id_ref),
                operator_name_two,
            )
            return
        self.single_contract_deletion_requested.emit(
            int(agent_id_ref),
            int(selected_agent_id),
            operator_name_two,
        )

    def keyPressEvent(self, event):
        """Process key press event to process the deletion of agents and contracts"""
        if event.key() == Qt.Key_Delete:
            selected_items = self.selectedItems()

            for item in selected_items:
                parent = item.parent()
                text_of_selected_item = self.currentItem().text(0)

                if not parent:  # -> check if item has no parent to trigger the deletion of all agents of a certain type
                    self.all_agents_deletion_requested.emit(text_of_selected_item)
                    return

                agent_id_ref = self._get_id_from_item_desc(text_of_selected_item)

                if agent_id_ref:
                    # item has a parent and is a contract or single agent
                    non_agent_root_item = self.currentItem()
                    print(non_agent_root_item)

                    if not self.currentItem().text(0).__contains__(
                            "("):  # use display id to check if item is a contract

                        self.agent_deletion_requested.emit(int(agent_id_ref))

                        return

                    self.process_contract_deletion(
                        text_of_selected_item, non_agent_root_item, agent_id_ref
                    )

                    return
